# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from six import iteritems, with_metaclass
from IPython.core.getipython import get_ipython
'''
Provide easy access to the Shell user variables
Sample use:
    ShellAccess.sc => access to the SparkContext
    ShellAccess["sqlContext"] => access the SQLContext
    ShellAccess.myVar = "Hello" => set a new variable called myVar in the user namespace
'''
class ShellAccess(with_metaclass( 
        type("", (type,), {
            "__getitem__":lambda cls, key: cls.do_get_item(key),
            "__setitem__":lambda cls, key, val: get_ipython().user_ns.update({key:val}),
            "__getattr__":lambda cls, key: cls.do_get_item(key),
            "__setattr__":lambda cls, key, val: get_ipython().user_ns.update({key:val}),
            "__delitem__":lambda cls, key: get_ipython().user_ns.pop(key, None),
            "__iter__": lambda cls: iter(get_ipython().user_ns.keys()),
            "keys": lambda cls: get_ipython().user_ns.keys()
        }), object
    )):

    @staticmethod
    def do_get_item(key):
        parts = key.split(".")
        obj = get_ipython().user_ns.get(parts[0])
        for p in parts[1:]:
            if obj is not None:
                obj = getattr(obj, p, None)
        return obj

    @staticmethod
    def update(**kwargs):
        for key, val in iteritems(kwargs):
            ShellAccess[key] = val



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.app import *
from pixiedust.utils import Logger
from pixiedust.display.datahandler import getDataHandler
from pixiedust.utils import *
from pixiedust.utils.dataFrameMisc import isPySparkDataFrame, isPandasDataFrame
from pixiedust.utils.environment import Environment

from pixiedust.utils.astParse import parse_function_call
from IPython.core.getipython import get_ipython

from pixiedust.utils.shellAccess import ShellAccess
from pixiedust.display.chart.options.baseOptions import BaseOptions
import json

@PixieApp
@Logger()
class FilterApp(BaseOptions):
    def get_custom_options(self):
        return { "runInDialog":"false" }

    def setup(self):
        BaseOptions.setup(self)

        self.filter_options = {}
        if self.run_options is not None and 'filter' in self.run_options:
            self.filter_options = json.loads(self.run_options['filter'])
    
        self.parent_prefix = self.parsed_command['kwargs']['prefix']
        self.fieldNamesAndTypes = self.get_field_names_and_types(True, True)
        self.fieldNames = self.get_field_names(True)

    def reset_data(self):
        self.df = self.parent_entity
        self.dfh = getDataHandler(None, self.df)

    def clear_filter(self, v):
        self.reset_data()
        self.filter_options = {}
        self.on_update()

    def get_new_options(self):
        return {
            "filter": json.dumps(self.filter_options)
        }

    @route()
    @templateArgs
    def main_screen(self): 
        self.reset_data()
        cleared = None
        cols = self.fieldNames
        filteredField = self.filter_options['field'] if 'field' in self.filter_options else ''

        if filteredField not in cols:
            filteredField = ''
            self.filter_options = {}
        
        return """
        <style>
        #columnselect{{prefix}}, #constraintsselect{{prefix}} { margin:0px; width:calc(100% - 10px); }
        .div-inline { display:inline-block;vertical-align:top } 
        .filter-heading {vertical-align:top;font-weight:500;margin-bottom:10px;}
        .filter-heading > span:last-of-type { font-size: small;font-weight: 300;display: inline-block;padding-left: 10px; }
        .form-inline select {vertical-align:top !important}
        .new-line {display:block !important} 
        .query-input { min-width: calc(100% - 10px); }
        input.form-control, .filter-clear { margin-right: 12px;}
        .stats .label {margin: 0 0.2em} 
        .filter-ui .label {font-weight:300} 
        .stats-table { width: 100%; } 
        .stats-table td { vertical-align:top; }
        .stats-table th { text-align:center; }
        .stats-table tr td:nth-child(2n) { text-align: left; }
        .panel-heading .data-toggle:before { font-family:fontAwesome; content:"\\f0da\\00a0\\00a0"; }
        .panel-heading .data-toggle[aria-expanded="true"]:before { font-family:fontAwesome; content:"\\f0d7\\00a0\\00a0"; }
        a.data-toggle, a.data-toggle:link, a.data-toggle:visited { text-decoration: none }
        #casematterscheck_{{prefix}}, regexcheck_{{prefix}} {margin: 0;}
        #df-stats-panel{{prefix}}, #regex-help-panel{{prefix}} { margin-bottom: 0; }
        #df-stats-panel{{prefix}} .panel-heading, #regex-help-panel{{prefix}} .panel-heading { font-size: 13px; padding: 9px 15px; }
        #df-stats-panel{{prefix}} .panel-body, #regex-help-panel{{prefix}} .panel-body { padding:0;max-height:240px;white-space:nowrap;overflow-y:scroll}
        #regex-help-panel{{prefix}} h3, #regex-help-panel{{prefix}} dt { margin-left: 10px}
        #regex-help-panel{{prefix}} dd { margin-left: 25px}
        </style>
        <div class="filter-ui">
            <div class="filter-heading panel-title">
                <span>Filter:</span>
                <span id="results{{prefix}}" class="no_loading_msg"></span>
            </div>
            
            <form class="form-inline row">
                <div class="form-group col-sm-2">
                    <select id="columnselect{{prefix}}" pd_options="field=$val(columnselect{{prefix}})" pd_target="constraints{{prefix}}" class="form-control filter-select" aria-label="select column">
                        <option value="--select-column--" disabled selected>Select a Field</option>
                    {%for col in cols %}
                        <option value="{{col}}">{{col}}</option>
                    {%endfor%}
                    </select>
                </div>
                <div class="form-group col-sm-10">
                    <div id="constraints{{prefix}}" class="no_loading_msg"></div>
                </div>
            </form>
        </div>
        <script>
        function filterInfo{{prefix}}() {
            var v = $('#manualvalue_{{prefix}}').val()
            if (v) {
                var r = $('#regexcheck_{{prefix}}').is(':checked') ? 'True' : 'False'
                var m = $('#casematterscheck_{{prefix}}').is(':checked') ? 'True' : 'False'
                var filtermsg = 'field: ' + $('#columnselect{{prefix}}').val() + ', constraint: ' + ($('#constraintsselect{{prefix}}').val() || 'None') + ', value: ' + v + ', casematters: ' + m + ', regex: ' + r
                $('#filterbutton{{this.parent_prefix}}').attr('title', 'Filter - ' + filtermsg)
                $('#results{{prefix}}').text(filtermsg)
                $('#filterbutton{{this.parent_prefix}}').css({
                    'background-color': 'orange',
                    'border-color': 'orange',
                    'color': 'white'
                })
            } else {
                clearFilterInfo{{prefix}}(true)
            }
        }
        function clearFilterInfo{{prefix}}(donotempty) {
            $('#filterbutton{{this.parent_prefix}}').attr('title', 'Filter')
            $('#results{{prefix}}').text('')
            $('#filterbutton{{this.parent_prefix}}').css({
                'background-color': '',
                'border-color': '',
                'color': ''
            })
            $('#manualvalue_{{prefix}}').val('')
            if (!donotempty) {
                $('#columnselect{{prefix}}').val('--select-column--')
                $('#constraints{{prefix}}').empty()
            }

            return ''
        }
        function valOnUpdate{{prefix}}() {
            filterInfo{{prefix}}()
            var v = $('#manualvalue_{{prefix}}').val()
            return v.replace(/\\\\/g, '\\\\\\\\')
        }
        if ('{{filteredField}}') {
            $('#columnselect{{prefix}}').val('{{filteredField}}').change()
        }
        </script>
        """
    
    @route(field="*")
    @templateArgs
    def colnamechange(self, field):
        # self.reset_data()
        filteredConstraint = self.filter_options['constraint'] if 'constraint' in self.filter_options else ''
        filteredValue = (self.filter_options['value'] if 'value' in self.filter_options else '').replace("\\", "\\\\")
        filteredRegex = self.filter_options['regex'] if 'regex' in self.filter_options else 'False'
        filteredCase = self.filter_options['case_matter'] if 'case_matter' in self.filter_options else 'False'

        isNumericField = self.dfh.isNumericField(field)
        if isNumericField and (filteredConstraint == 'None' or not filteredConstraint):
            filteredConstraint = 'equal_to'

        stats = ""
        controls = ""
        manualvalue = """
        <div class="form-group col-sm-4">
            <input class="form-control query-input" id="manualvalue_{{prefix}}" placeholder="Enter value">
        </div>
        """
        script = """
            $('#manualvalue_{{prefix}}').keydown(function(event){
                if (event.keyCode == 13) {
                    event.preventDefault()
                    return false
                }
            })
            if ('{{filteredConstraint}}') {
                $('#constraintsselect{{prefix}}').val('{{filteredConstraint}}')
            }
            if ('{{filteredValue}}') {
                $('#manualvalue_{{prefix}}').val('{{filteredValue}}')
            }
        """

        if not isNumericField:
            script += """
                if ('{{filteredRegex}}') {
                    $('#regexcheck_{{prefix}}').prop('checked', ('{{filteredRegex}}'.toLowerCase() === 'true'));
                }
                if ('{{filteredCase}}') {
                    $('#casematterscheck_{{prefix}}').prop('checked', ('{{filteredCase}}'.toLowerCase() === 'true'));
                }
            """
            controls = """
            <div class="form-group col-sm-2">
                <input type="checkbox" id="casematterscheck_{{prefix}}" value="casematters"> Case-sensitive<br/>
                <input type="checkbox" id="regexcheck_{{prefix}}" value="regex"> Regex
            </div>
              """
            stats += """
            <div class="form-group col-sm-4">
                <div id="regex-help-panel{{prefix}}" class="panel panel-default">
                    <div class="panel-heading">
                        <h4 class="panel-title " style="margin:0px">
                            <a data-toggle="collapse" class="data-toggle" href="#regex-help-{{prefix}}">Regex help</a>
                        </h4>
                    </div>
                    <div id="regex-help-{{prefix}}" class="panel-collapse collapse">
                        <div class="panel-body">
                            <h3>Character classes</h3>
                            <dt>[abc]</dt><dd>matches a or b, or c.</dd>
                            <dt>[^abc]</dt><dd>negation, matches everything except a, b, or c.</dd>
                            <dt>[a-c]</dt><dd>range, matches a or b, or c.</dd>
                            <dt>[a-c[f-h]]</dt><dd>union, matches a, b, c, f, g, h.</dd>
                            <dt>[a-c&&[b-c]]</dt><dd>intersection, matches b or c.</dd>
                            <dt>[a-c&&[^b-c]]</dt><dd>subtraction, matches a.</dd>

                            <h3>Predefined character classes</h3>
                            <dt>.</dt><dd>Any character.</dd>
                            <dt>&#92;d</dt><dd>A digit: [0-9]</dd>
                            <dt>&#92;D</dt><dd>A non-digit: [^0-9]</dd>
                            <dt>&#92;s</dt><dd>A whitespace character: [ &#92;t&#92;n&#92;x0B&#92;f&#92;r]</dd>
                            <dt>&#92;S</dt><dd>A non-whitespace character: [^&#92;s]</dd>
                            <dt>&#92;w</dt><dd>A word character: [a-zA-Z_0-9]</dd>
                            <dt>&#92;W</dt><dd>A non-word character: [^&#92;w]</dd>

                            <h3>Boundary matches</h3>
                            <dt>^</dt><dd>The beginning of a line.</dd>
                            <dt>$</dt><dd>The end of a line.</dd>
                            <dt>&#92;b</dt><dd>A word boundary.</dd>
                            <dt>&#92;B</dt><dd>A non-word boundary.</dd>
                            <dt>&#92;A</dt><dd>The beginning of the input.</dd>
                            <dt>&#92;G</dt><dd>The end of the previous match.</dd>
                            <dt>&#92;Z</dt><dd>The end of the input but for the final terminator, if any.</dd>
                            <dt>&#92;z</dt><dd>The end of the input.</dd>
                        </div>
                    </div>
                </div>
            </div>
            """
        
        else: # numeric field
            controls = """
            <div class="form-group col-sm-1">
            <select id="constraintsselect{{prefix}}" class="form-control filter-select col-3">
                <option value="less_than"> < </option>
                <option value="equal_to"> = </option>
                <option value="greater_than"> > </option>
            </select>
            </div>
            """

            # GENERATE A TABLE OF STATISTICS
            stats += """
            <div class="form-group col-sm-5">
                <div id="df-stats-panel{{prefix}}" class="panel panel-default">
                    <div class="panel-heading">
                        <h4 class="panel-title " style="margin:0px">
                            <a data-toggle="collapse" class="data-toggle" href="#df-stats-{{prefix}}">Statistics: <span style="font-weight: normal">{{field}}</span></a>
                        </h4>
                    </div>
                    <div id="df-stats-{{prefix}}" class="panel-collapse collapse">
                        <div class="panel-body">
            """

            stats += self.stats_table(field)

            stats += """
                        </div>
                    </div>
                </div>
            </div>
            """

        script += """
            filterInfo{{prefix}}()
        """
        submit = """
        <div class="form-group col-sm-1">
            <button type="button" class="btn btn-default btn-primary filter-submit" pd_options="field={{field}};constraint=None;val=$val(valOnUpdate{{prefix}});casematters=$val(casematterscheck_{{prefix}});regex=$val(regexcheck_{{prefix}})" pd_target="results{{prefix}}">Apply</button>
        </div>
        """
        if isNumericField:
            submit = """
            <div class="form-group col-sm-1">
                <button type="button" class="btn btn-default btn-primary filter-submit" pd_options="field={{field}};constraint=$val(constraintsselect{{prefix}});val=$val(valOnUpdate{{prefix}});casematters=False;regex=False" pd_target="results{{prefix}}">Apply</button>
            </div>
            """
        
        clear = """
        <div class="form-group col-sm-1">
            <button type="button" class="btn btn-default filter-clear" pd_script="self.clear_filter('$val(clearFilterInfo{{prefix}})')">Clear</button>
        </div>"""

        if not isNumericField:
            return '<div class="row">' + manualvalue + controls + submit + clear + stats + '<script>' + script + '</script></div>'
        else:
            return '<div class="row">' + controls + manualvalue + submit + clear + stats + '<script>' + script + '</script></div>'

    @route(field="*", constraint="*", casematters="*")
    def noqueryvalue(self, field, constraint, casematters): 
        # called when user submits without entering anything in the query <input>
        return self.compute(field, constraint, '', casematters, 'false')

    @route(field="*", constraint="*", val="*", casematters="*", regex="*")
    def compute(self, field, constraint, val, casematters, regex):
        isNumericField = self.dfh.isNumericField(field)
        if not val or not val.strip() or val.strip() == 'None':
            val = 'None'

        if isNumericField and constraint not in ['less_than', 'greater_than', 'equal_to']:
            return "<h4>A constraint is required</h4>"
        elif isNumericField and val != 'None' and not val.replace('.','',1).replace('-','',1).isdigit():
            return "<h4>A numeric value is required</h4>"
        else:
            self.filter_options = {
                "field": field,
                "constraint": constraint,
                "value": val.replace("\\", "\\\\"),
                "case_matter": casematters,
                "regex": regex
            }

            self.on_update()
            return "field: {}, constraint: {}, value: {}, casematters: {}, regex: {}".format(field, constraint, val, casematters, regex)

    def stats_table(self, field):
        self.summary_stats = []
        self.quantiles = []
        self.frequents = []
        
        if isPySparkDataFrame(self.df):
            statsdf = self.df.describe(field)
            lbls = ['count','mean','std','min','max']

            for i in [0,1,2,3,4]:
                if i == 0:
                    self.summary_stats.append((lbls[i], "{:.0f}".format(float(statsdf.collect()[i][1]))))
                else:
                    self.summary_stats.append((lbls[i], "{:.2f}".format(float(statsdf.collect()[i][1]))))
                    
            if Environment.sparkVersion == 2:
                lbls = ['2%','9%','25%','50%','75%','91%','98%']
                quants = self.df.approxQuantile(field, [.02, .09, .25, .50, .75, .91, .98], 0.1)
                for i, q in enumerate(quants):
                    self.quantiles.append((lbls[i] + "ile", "{:.2f}".format(q)))
                
            freqdf = self.df.stat.freqItems([field], 0.1)
            freqlist = freqdf.collect()[0][field+'_freqItems']
            stop = 5
            for i in freqlist:
                if stop > 0:
                    self.frequents.append(str(i))
                stop = stop - 1
        else:
            if not isPandasDataFrame(self.df):
                self.df = self.data_handler.entity
            if isPandasDataFrame(self.df):
                statsdf = self.df[field].describe([.02, .09, .25, .50, .75, .91, .98])
                lbls = ['count','mean','std','min','max']

                for i in range(0,len(lbls)):
                    if i == 0:
                        self.summary_stats.append((lbls[i], "{:.0f}".format(statsdf[i])))
                    else:
                        self.summary_stats.append((lbls[i], "{:.2f}".format(statsdf[lbls[i]])))

                lbls = ['2%','9%','25%','50%','75%','91%','98%']
                for i in range(0,len(lbls)):
                    self.quantiles.append((lbls[i] + "ile", "{:.2f}".format(statsdf[lbls[i]])))

                freqseries = self.df[field].value_counts()
                stop = 5
                for ix in freqseries.index:
                    if stop > 0:
                        self.frequents.append(str(ix))
                    stop = stop - 1
                
        summaryname = '<br>'.join(s[0] for s in self.summary_stats)
        summaryvalue = '<br>'.join(s[1] for s in self.summary_stats)
        quantname = '<br>'.join(q[0] for q in self.quantiles)
        quantvalue = '<br>'.join(q[1] for q in self.quantiles)
        freqvalue = '<br>'.join(f for f in self.frequents)

        table = """
            <table class="stats-table">
                <thead>
                    <tr>""" 
        table += "<th colspan='2'>Summary</th> <th colspan='2'>Quantiles</th> <th>Frequents</th>" if Environment.sparkVersion == 2 else "<th colspan='2'>Summary</th> <th>Frequents</th>"
        table += """
                    </tr>
                </thead>
                <tbody>
                    <tr>"""
        table += "<td>{}</td> <td>{}</td> <td>{}</td> <td>{}</td> <td>{}</td>" if Environment.sparkVersion == 2 else "<td>{}</td> <td>{}</td> <td>{}</td>"
        table += """
                    </tr>
                </tbody>
            </table>"""
            
        if Environment.sparkVersion == 2:
            return table.format(summaryname, summaryvalue, quantname, quantvalue, freqvalue)
        else:
            return table.format(summaryname, summaryvalue, freqvalue)

    def on_update(self):
        return self.on_ok(avoid_metadata=False, override_keys=['filter'])



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from ..bokeh import BokehStreamingDisplay
import numpy as np

N=100
class ScatterPlotStreamingDisplay(BokehStreamingDisplay):
    def createGlyphRenderer(self, figure, x, y):
        self.radii = np.random.random(size=N) * 2
        self.colors = ["#%02x%02x%02x" % (int(r), int(g), 150) for r, g in zip(50+2*self._toNPArray(x), 30+2*self._toNPArray(y))]
        self.i = 0
        return figure.circle(x,y, radius=self.radii, 
                fill_color=self.colors, fill_alpha=0.6, line_color=None, 
                hover_fill_color="black", hover_fill_alpha=0.7, hover_line_color=None)
        
    def updateGlyphRenderer(self, figure, glyphRenderer):
        self.i +=1 
        figure.title.text = str(self.i)    
        glyphRenderer.data_source.data['radius'] = self.radii * (2 + np.sin(self.i/5))


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.app import *
from pixiedust.utils import Logger
from pixiedust.display.chart.options.defaultOptions import DefaultOptions
from pixiedust.display.chart.options.optionsShell import OptionsShell
from pixiedust.display.chart.options.components.BinQuantiles import BinQuantiles

@Logger()
class MapboxAccessToken(object):
    @route(widget="mapboxAccessToken")
    def mapbox_access_token_widget(self, optid, mapboxtoken):
        return """
<div class="form-group">
  <label for="mapboxoption{{optid}}{{prefix}}">
    <a href="https://www.mapbox.com/help/create-api-access-token/" target="_mapboxwin">Mapbox Access Token</a>: <i class="fa fa-question-circle" title="Get a Mapbox access token by creating a free account at mapbox.com"></i>
  </label>
  <input type="text" class="form-control" id="mapboxoption{{optid}}{{prefix}}" name="{{optid}}" value="{{mapboxtoken}}"
      pd_script="self.options_callback('{{optid}}', '$val(mapboxoption{{optid}}{{prefix}})')" onkeyup="$(this).trigger('click');">
</div>
"""

class MapboxCustomBaseColor(object):
    @route(widget="mapboxCustomBaseColor")
    def mapbox_color_picker_widget(self, optid, custombasecolor, labelname="Custom Base Color:"):
        return """
<div class="form-group">
<label for="mapboxoption{{optid}}{{prefix}}">{{labelname}}</label>
<input type="color" class="form-control" id="mapboxoption{{optid}}{{prefix}}" name="{{optid}}" value="{{custombasecolor}}"
  pd_script="self.options_callback('{{optid}}', '$val(mapboxoption{{optid}}{{prefix}})')" onkeyup="$(this).trigger('click');">
</div>
"""

@PixieApp
@Logger()
class MapboxOptions(DefaultOptions, MapboxAccessToken, MapboxCustomBaseColor):

    mapbox_default_token = "pk.eyJ1IjoibWFwYm94IiwiYSI6ImNpejY4M29iazA2Z2gycXA4N2pmbDZmangifQ.-g_vE53SD2WrJ6tFX7QHmA"

    def setup(self):
        DefaultOptions.setup(self)

        self.chart_options.append({
            "optid": "mapboxtoken",
            "classname": "field-width-50",
            "mapboxtoken": lambda: self.run_options.get("mapboxtoken") or self.mapbox_default_token,
            "widget": "mapboxAccessToken"
        })

        self.chart_options.append({
            "optid": "custombasecolor",
            "classname": "field-width-50",
            "custombasecolor": lambda: self.run_options.get("custombasecolor") or "#ff0000",
            "labelname": "Custom Base Color:",
            "widget": "mapboxCustomBaseColor"
        })

        self.chart_options.append({
            "optid": "custombasecolorsecondary",
            "classname": "field-width-50",
            "custombasecolor": lambda: self.run_options.get("custombasecolorsecondary") or "#ff0000",
            "labelname": "Secondary Custom Base Color:",
            "widget": "mapboxCustomBaseColor"
        })

        self.new_options["mapboxtoken"] = self.run_options.get("mapboxtoken") or self.mapbox_default_token

    def key_fields_type(self):
        return ['string', 'numeric']

    def value_fields_type(self):
        return ['any']

@PixieApp
@Logger()
class NumBinsOptions(OptionsShell, BinQuantiles):

    def setup(self):
        OptionsShell.setup(self)
        # should only be one entry in the chart option list for chart name
        # don't want that, just get rid of it
        self.chart_options.pop()

        numbins = int(self.run_options.get("numbins") or 5)
        self.setupQuantiles(numbins)

        self.chart_options.append({
            "optid": "binRanges",
            "classname": "field-width-100 no_loading_msg",
            "numbins": numbins,
            "quantiles": self.new_options["quantiles"],
            "widget": "pdBinQuantiles"
        })

    # default equal size quantiles
    # for numBins=5, quantiles are 0, 0.25, 0.50, 0.75, 1
    def generateQuantiles(self, numBins):
        quantiles = []
        for i in range(numBins):
            quantiles.append(float(i)/(numBins-1.0))
        return ",".join(map(str,quantiles))

    # set up quantiles based on numbins/presence of quantiles string in metadata
    def setupQuantiles(self, numBins):
        if self.run_options.get("quantiles"):
            if len(self.run_options["quantiles"].split(",")) == numBins:
                self.new_options["quantiles"] = self.run_options["quantiles"]
                return

        self.new_options["quantiles"] = self.generateQuantiles(numBins)
        return



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.app import *
from pixiedust.utils import Logger
from pixiedust.display.chart.options.defaultOptions import DefaultOptions

@Logger()
class GoogleMapApiKey(object):
    @route(widget="googleMapApiKey")
    def google_map_api_key(self, optid, googlemapapikey):
        return """
<div class="form-group">
  <label for="googlemapoption{{optid}}{{prefix}}">
    <a href="https://support.google.com/googleapi/answer/6158862?hl=en&ref_topic=7013279" target="_blank">Google Maps API Key</a>: <i class="fa fa-question-circle" title="Get a Google Maps API key by accessing the Google API Console"></i>
  </label>
  <input type="text" class="form-control" id="googlemapoption{{optid}}{{prefix}}" name="{{optid}}" value="{{googlemapapikey}}"
    pd_script="self.options_callback('{{optid}}', '$val(googlemapoption{{optid}}{{prefix}})')" onkeyup="$(this).trigger('click');">
</div>
"""

@PixieApp
@Logger()
class GoogleMapOptions(DefaultOptions, GoogleMapApiKey):
    def setup(self):
        DefaultOptions.setup(self)

        self.chart_options.append({
            "optid": "googlemapapikey",
            "classname": "field-width-50",
            "googlemapapikey": lambda: self.options.get("googlemapapikey") or "",
            "widget": "googleMapApiKey"
        })

    def key_fields_type(self):
        return ['string', 'numeric']

    def value_fields_type(self):
        return ['any']



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.display import CellHandshake
from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.utils import Logger,cache
from ..baseChartDisplay import BaseChartDisplay
from six import with_metaclass
from abc import abstractmethod, ABCMeta
from bokeh.plotting import figure, output_notebook
from bokeh.models.tools import *
import pkg_resources
from bokeh.palettes import viridis, magma, plasma, linear_palette, Spectral9, d3

try:
    from bokeh.embed import components as notebook_div
except ImportError:
    from bokeh.io import notebook_div

@PixiedustRenderer(rendererId="bokeh")
@Logger()
class BokehBaseDisplay(with_metaclass(ABCMeta, BaseChartDisplay)):
    CellHandshake.addCallbackSniffer( lambda: "{'nostore_bokeh':!!window.Bokeh}")

    #get the bokeh version
    try:
        bokeh_version = pkg_resources.get_distribution("bokeh").parsed_version._version.release
    except:
        bokeh_version = None
        self.exception("Unable to get bokeh version")

    def __init__(self, options, entity, dataHandler=None):
        super(BokehBaseDisplay,self).__init__(options,entity,dataHandler)
        #no support for orientation
        self.no_orientation = True
    
    """
    Default implementation for creating a chart object.
    """
    def createBokehChart(self):
        return figure()

    def getPreferredOutputWidth(self):
        return super(BokehBaseDisplay,self).getPreferredOutputWidth() * 0.92

    def get_common_figure_options(self):
        options = {}
        x_fields = self.getKeyFields()
        if len(x_fields) == 1 and (
            self.options.get("timeseries", 'false') == 'true' or self.dataHandler.isDateField(x_fields[0])
            ):
            options["x_axis_type"] = "datetime"
        elif self.getBooleanOption("logx", False):
            options["x_axis_type"] = "log"

        if self.getBooleanOption("logy", False):
            options["y_axis_type"] = "log"        
        return options

    @cache(fieldName="_loadJS")
    def getLoadJS(self):
        loadJS = self._load_notebook_html(hide_banner=True)
        return loadJS

    def colorPalette(self, size=None):
        # https://bokeh.pydata.org/en/latest/docs/reference/palettes.html
        # https://bokeh.pydata.org/en/latest/docs/reference/colors.html
        color = list(d3['Category10'][10]) # [ 'orangered', 'cornflowerblue',  ]
        # color = list(Spectral9).reverse()
        if size is None:
            return color
        elif size <= len(color):
            return color[0:size]
        elif size <= 256:
            return linear_palette(plasma(256), size)
        else:
            return linear_palette(plasma(256) + viridis(256) + magma(256), size)

    def doRenderChart(self):        
        def genMarkup(chartFigure):
            s = chartFigure[0] if isinstance(chartFigure, tuple) else chartFigure
            d = chartFigure[1] if isinstance(chartFigure, tuple) else ''
            return self.env.from_string("""
                    <script class="pd_save">
                    function setChartScript() {{
                        if (!window.Bokeh) {{
                            setTimeout(setChartScript, 250)
                        }} else {{
                            var d = document.getElementById("pd-bkchartdiv-{p}")
                            if (d){{
                                var el = document.createElement('div')
                                el.innerHTML = `{chartScript}`
                                var chartscript = el.childNodes[1]
                                var s = document.createElement("script")
                                s.innerHTML = chartscript.innerHTML
                                d.parentNode.insertBefore(s, d)
                            }}
                        }}
                    }}
                    if (!window.Bokeh && !window.autoload){{
                        window.autoload=true;
                        {loadJS}
                    }}
                    setChartScript()
                    </script>
                    <div style="padding:5px" id="pd-bkchartdiv-{p}">{chartDiv}</div>
                    {{%for message in messages%}}
                        <div>{{{{message}}}}</div>
                    {{%endfor%}}
                """.format(chartScript=s.replace('</script>', '<\/script>'), chartDiv=d, loadJS=self.getLoadJS(), p=self.getPrefix())
            ).render(messages=self.messages)

        minBkVer = (0,12,9)
        if BokehBaseDisplay.bokeh_version < minBkVer:
            raise Exception("""
                <div>Incorrect version of Bokeh detected. Expected {0} or greater, got {1}</div>
                <div>Please upgrade by using the following command: <b>!pip install --user --upgrade bokeh</b></div>
            """.format(minBkVer, BokehBaseDisplay.bokeh_version))
        clientHasBokeh = self.options.get("nostore_bokeh", "false") == "true"
        if not clientHasBokeh:          
            output_notebook(hide_banner=True)
        charts = self.createBokehChart()

        if not isinstance(charts, list):
            # charts.add_tools(ResizeTool())
            #bokeh 0.12.5 has a non backward compatible change on the title field. It is now of type Title
            #following line is making sure that we are still working with 0.12.4 and below
            if hasattr(charts, "title") and hasattr(charts.title, "text"):
                charts.title.text = self.options.get("title", "")
            else:
                charts.title = self.options.get("title", "")
            charts.plot_width = int(self.getPreferredOutputWidth() - 10 )
            charts.plot_height = int(self.getPreferredOutputHeight() - 10  )
            charts.grid.grid_line_alpha=0.3
            return genMarkup(notebook_div(charts))
        else:
            from bokeh.layouts import gridplot
            ncols = 2
            nrows = len(charts)/2 + len(charts)%2
            
            w = self.getPreferredOutputWidth()/ncols if len(charts) > 1 else self.getPreferredOutputWidth()
            h = w * self.getHeightWidthRatio() if len(charts) > 1 else self.getPreferredOutputHeight()
            for chart in charts:
                chart.plot_width = int(w - 5)
                chart.plot_height = int (h - 5)

            return genMarkup(notebook_div(gridplot(charts, ncols=ncols)))


    # no longer part of bokeh
    #  https://github.com/bokeh/bokeh/pull/6928#issuecomment-329040653
    #  https://www.bvbcode.com/code/9xhgwcsf-2674609
    def _load_notebook_html(self, resources=None, hide_banner=False, load_timeout=5000):
        from bokeh.core.templates import AUTOLOAD_NB_JS
        from bokeh.util.serialization import make_id
        from bokeh.resources import CDN
    
        if resources is None:
            resources = CDN
    
        element_id = make_id()
    
        js = AUTOLOAD_NB_JS.render(
            elementid = '' if hide_banner else element_id,
            js_urls  = resources.js_files,
            css_urls = resources.css_files,
            js_raw   = resources.js_raw,
            css_raw  = resources.css_raw_str,
            force    = 1,
            timeout  = load_timeout
        )
    
        return js


from pixiedust.display.chart.renderers import PixiedustRenderer
from .TableBaseDisplay import TableBaseDisplay
from pixiedust.utils import cache
from pixiedust.utils import Logger

@PixiedustRenderer(id="tableView")
@Logger()
class TableDisplay(TableBaseDisplay):

    def doRenderChart(self):
        wpdf = self.getWorkingPandasDataFrame()
        table_showrows = self.options.get("table_showrows", 'All')
        if table_showrows == 'Missing values':
            wpdf = wpdf[wpdf.isna().any(axis=1) if hasattr(wpdf,'isna') else wpdf.isnull().any(axis=1)]
        elif table_showrows == 'Not missing values':
            wpdf = wpdf[wpdf.notna().all(axis=1) if hasattr(wpdf,'notna') else wpdf.notnull().all(axis=1)]


        return self.renderTemplate("table.html", wpdf=wpdf,
            table_noschema=self.options.get("table_noschema", "false"),
            table_nocount=self.options.get("table_nocount", "false"),
            table_nosearch=self.options.get("table_nosearch", "false"))



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.streamingDisplay import StreamingDisplay
from pixiedust.display.display import CellHandshake
import pandas
import numpy as np
from bokeh.io import push_notebook, output_notebook
from bokeh.io.state import curstate
from bokeh.io.notebook import show_doc
from bokeh.models import HoverTool
from bokeh.plotting import figure
from pixiedust.utils import Logger

@Logger()
class BokehStreamingDisplay(StreamingDisplay):
    CellHandshake.addCallbackSniffer( lambda: "{'nostore_bokeh':!!window.Bokeh}")

    def __init__(self, options, entity, dataHandler=None):
        super(BokehStreamingDisplay,self).__init__(options,entity,dataHandler)
        self.TOOLS="crosshair,pan,wheel_zoom,box_zoom,reset,tap,box_select,lasso_select"
        self.figure = figure(tools=self.TOOLS)
        self.figure.axis.major_label_text_font_size = "18pt"
        self.hover = HoverTool(tooltips=None, mode="vline")
        self.figure.add_tools(self.hover)

        self.comms_handle = None
        self.glyphRenderer = None
        self.setup();

    def setup(self):
        pass

    def createGlyphRenderer(self, figure):
        return None

    def updateGlyphRenderer(self, figure, glyphRenderer):
        pass

    def _concatArrays(self,a,b):
        if type(a) != type(b):
            raise Exception("Can't concatenate objects of different types")
        if isinstance(a, list):
            return a+b
        elif isinstance(a, np.ndarray):
            return np.concatenate((a,b))
        raise Exception("Can't concatenate: unsupported types")

    def _delWindowElements(self, array):
        if isinstance(array, list):
            del array[:len(array)-self.windowSize]
            return array
        elif isinstance(array, np.ndarray):
            array = np.delete(array, range(0, len(array) - self.windowSize))
            return array
        raise Exception("Can't delete: unsupported type")

    def _toNPArray(self, a ):
        if isinstance(a, list):
            return np.array(a)
        elif isinstance(a, np.ndarray):
            return a
        raise Exception("Can't cast to np array: unsupported type")

    def doRender(self, handlerId):
        clientHasBokeh = self.options.get("nostore_bokeh", "false") == "true"
        if not clientHasBokeh:
            output_notebook(hide_banner=True)
        data = self.entity.getNextData()
        if data is None:
            return

        x = None
        y = None

        if isinstance(data, (list,np.ndarray)):
            x = list(range(self.windowSize)) if self.glyphRenderer is None else self.glyphRenderer.data_source.data['x']
            y = data if self.glyphRenderer is None else self._concatArrays(self.glyphRenderer.data_source.data['y'],data)
            if len(y) < self.windowSize:
                y = [0]*(self.windowSize-len(y)) + y
            elif len(y) > self.windowSize:
                y = self._delWindowElements(y)
        elif isinstance(data, pandas.core.frame.DataFrame):
            pd = pd.drop(pd.index[[0]])
            #pd.index = list(range(len(pd.index)))
            pd['x'] = list(range(len(pd.index)))
        else:
            x = data[0]
            y = data[1]

        if self.glyphRenderer is None:
            self.glyphRenderer = self.createGlyphRenderer( self.figure, x, y )
        else:
            self.updateGlyphRenderer( self.figure, self.glyphRenderer)

        if self.glyphRenderer is None:
            print("Error: no glyphRenderer found")
            return

        self.glyphRenderer.data_source.data['x'] = x
        self.glyphRenderer.data_source.data['y'] = y

        if not self.comms_handle:
            state = curstate()
            doc = state.document
            if self.figure not in doc.roots:
               doc.add_root(self.figure)
            self.comms_handle = show_doc(self.figure, state, notebook_handle=True)
        else:
            push_notebook(handle = self.comms_handle)


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.utils import Logger
from .brunelBaseDisplay import BrunelBaseDisplay

@PixiedustRenderer(id="scatterPlot")
@Logger()
class ScatterPlotRenderer(BrunelBaseDisplay):

    def compute_brunel_magic(self):
        parts = ["point"]
        valueFields = self.getValueFields()
        keyFields = self.getKeyFields()
        
        parts.append("x({})".format(keyFields[0]))
        if len(valueFields) == 1:
            parts.append("y({})".format(valueFields[0]))
        elif len(valueFields) > 1:
            parts.append("y({})".format(",".join(valueFields)))
            parts.append("color(#series)")
        
        parts.append(self.get_sort())

        return parts



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from six import iteritems
import pixiedust
import ast
import zipfile
from pixiedust.utils.shellAccess import ShellAccess
from pixiedust.utils.template import PixiedustTemplateEnvironment
from pixiedust.utils.environment import Environment,scalaGateway
import pandas as pd
import uuid
import tempfile
from collections import OrderedDict
from IPython.display import display, HTML, Javascript
import json
import requests
from pandas.io.json import json_normalize
import os
try:
    from urllib.request import Request, urlopen, URLError, HTTPError
except ImportError:
    from urllib2 import Request, urlopen, URLError, HTTPError

dataDefs = OrderedDict([
    ("1", {
        "displayName": "Car performance data",
        "url": "https://github.com/ibm-watson-data-lab/open-data/raw/master/cars/cars.csv",
        "topic": "Transportation",
        "publisher": "IBM",
        "schema2": [('mpg','int'),('cylinders','int'),('engine','double'),('horsepower','int'),('weight','int'),
            ('acceleration','double'),('year','int'),('origin','string'),('name','string')]
    }),
    ("2", {
        "displayName": "Sample retail sales transactions, January 2009",
        "url": "https://raw.githubusercontent.com/ibm-watson-data-lab/open-data/master/salesjan2009/salesjan2009.csv",
        "topic": "Economy & Business",
        "publisher": "IBM Cloud Data Services"
    }),
    ("3", {
        "displayName": "Total population by country",
        "url": "https://apsportal.ibm.com/exchange-api/v1/entries/889ca053a19986a4445839358a91963e/data?accessKey=657b130d504ab539947e51b50f0e338e",
        "topic": "Society",
        "publisher": "IBM Cloud Data Services"
    }),
    ("4", {
        "displayName": "GoSales Transactions for Naive Bayes Model",
        "url": "https://apsportal.ibm.com/exchange-api/v1/entries/8044492073eb964f46597b4be06ff5ea/data?accessKey=bec2ed69d9c84bed53826348cdc5690b",
        "topic": "Leisure",
        "publisher": "IBM"
    }),
    ("5", {
        "displayName": "Election results by County",
        "url": "https://raw.githubusercontent.com/ibm-watson-data-lab/open-data/master/election2016/county_election_results.csv",
        "topic": "Society",
        "publisher": "IBM"
    }),
    ("6", {
        "displayName": "Million dollar home sales in Massachusetts, USA Feb 2017 through Jan 2018",
        "url": "https://raw.githubusercontent.com/ibm-watson-data-lab/open-data/master/homesales/milliondollarhomes.csv",
        "topic": "Economy & Business",
        "publisher": "Redfin.com"
    }),
    ("7", {
        "displayName": "Boston Crime data, 2-week sample",
        "url": "https://raw.githubusercontent.com/ibm-watson-data-lab/open-data/master/crime/boston_crime_sample.csv",
        "topic": "Society",
        "publisher": "City of Boston"
    })
])

@scalaGateway
def sampleData(dataId=None, type='csv', forcePandas=False, delimiter=','):
    global dataDefs
    return SampleData(dataDefs, forcePandas, delimiter).sampleData(dataId, type)

class SampleData(object):
    env = PixiedustTemplateEnvironment()
    def __init__(self, dataDefs, forcePandas, delimiter):
        self.dataDefs = dataDefs
        self.forcePandas = forcePandas
        self.url = ""
        self.delimiter = delimiter

    def sampleData(self, dataId = None, type='csv'):
        if dataId is None:
            self.printSampleDataList()
        elif str(dataId) in dataDefs:
            return self.loadSparkDataFrameFromSampleData(dataDefs[str(dataId)])
        elif "https://" in str(dataId) or "http://" in str(dataId) or "file://" in str(dataId):
            if type is 'json':
                self.url = str(dataId)
                return self.JSONloadSparkDataFrameFromUrl(str(dataId))
            else:
                return self.loadSparkDataFrameFromUrl(str(dataId))
        else:
            print("Unknown sample data identifier. Please choose an id from the list below")
            self.printSampleDataList()

    def printSampleDataList(self):
        display( HTML( self.env.getTemplate("sampleData.html").render( dataDefs = iteritems(self.dataDefs) ) ))

    def dataLoader(self, path, schema=None):
        if schema is not None and Environment.hasSpark:
            from pyspark.sql.types import StructType,StructField,IntegerType,DoubleType,StringType
            def getType(t):
                if t == 'int':
                    return IntegerType()
                elif t == 'double':
                    return DoubleType()
                else:
                    return StringType()

        if Environment.hasSpark and not self.forcePandas:
            if Environment.sparkVersion == 1:
                print("Loading file using 'com.databricks.spark.csv'")
                load = ShellAccess.sqlContext.read.format('com.databricks.spark.csv')
                if schema is not None:
                    return load.options(header='true', mode="DROPMALFORMED", delimiter=self.delimiter).load(path, schema=StructType([StructField(item[0], getType(item[1]), True) for item in schema]))
                else:
                    return load.options(header='true', mode="DROPMALFORMED", inferschema='true', delimiter=self.delimiter).load(path)
            else:
                print("Loading file using 'SparkSession'")
                csvload = ShellAccess.SparkSession.builder.getOrCreate() \
                    .read \
                    .format("org.apache.spark.sql.execution.datasources.csv.CSVFileFormat") \
                    .option("header", "true") \
                    .option("sep", self.delimiter) \
                    .option("mode", "DROPMALFORMED")
                if schema is not None:
                    return csvload.schema(StructType([StructField(item[0], getType(item[1]), True) for item in schema])).load(path)
                else:
                    return csvload.option("inferSchema", "true").load(path)
        else:
            print("Loading file using 'pandas'")
            try:
                return pd.read_csv(path, sep=self.delimiter)
            except UnicodeDecodeError:
                #Try ISO-8859-1
                return pd.read_csv(path, sep=self.delimiter, encoding = "ISO-8859-1")

    def JSONdataLoader(self, path, schema=None):
        if schema is not None and Environment.hasSpark:
            from pyspark.sql.types import StructType,StructField,IntegerType,DoubleType,StringType
            def getType(t):
                if t == 'int':
                    return IntegerType()
                elif t == 'double':
                    return DoubleType()
                else:
                    return StringType()

        res = open(path, 'r').read()
        if Environment.hasSpark and not self.forcePandas:
            if Environment.sparkVersion == 1:
                print("Loading file using a pySpark DataFrame for Spark 1")
                dataRDD = ShellAccess.sc.parallelize([res])
                return ShellAccess.sqlContext.jsonRDD(dataRDD)
            else:
                print("Loading file using a pySpark DataFrame for Spark 2")
                dataRDD = ShellAccess.sc.parallelize([res])
                return ShellAccess.spark.read.json(dataRDD)
        else:
            print("Loading file using 'pandas'")
            def load_json(text, stop = False):
                try:
                    return json.loads(text)
                except json.JSONDecodeError:
                    try:
                        return ast.literal_eval(text)
                    except SyntaxError:
                        if stop:
                            raise
                        return load_json("[" + ",".join(text.strip().split("\n")) + "]", True)

            data = load_json(res)
            df = json_normalize(data)
            return df

    def loadSparkDataFrameFromSampleData(self, dataDef):
        return Downloader(dataDef, self.forcePandas).download(self.dataLoader)

    def loadSparkDataFrameFromUrl(self, dataUrl):
        i = dataUrl.rfind('/')
        dataName = dataUrl[(i+1):]
        dataDef = {
            "displayName": dataUrl,
            "url": dataUrl
        }

        return Downloader(dataDef, self.forcePandas).download(self.dataLoader)

    def JSONloadSparkDataFrameFromUrl(self, dataUrl):
        i = dataUrl.rfind('/')
        dataName = dataUrl[(i+1):]
        dataDef = {
            "displayName": dataUrl,
            "url": dataUrl
        }

        return Downloader(dataDef, self.forcePandas).download(self.JSONdataLoader)

# Use of progress Monitor doesn't render correctly when previewed a saved notebook, turning it off until solution is found
useProgressMonitor = False
class Downloader(object):
    def __init__(self, dataDef, forcePandas):
        self.dataDef = dataDef
        self.forcePandas = forcePandas
        self.headers = {"User-Agent": "PixieDust Sample Data Downloader/1.0"}
        self.prefix = str(uuid.uuid4())[:8]

    def download(self, dataLoader):
        displayName = self.dataDef["displayName"]
        bytesDownloaded = 0
        if "path" in self.dataDef:
            path = self.dataDef["path"]
        else:
            url = self.dataDef["url"]
            req = Request(url, None, self.headers)
            print("Downloading '{0}' from {1}".format(displayName, url))
            tdir = '/home/spark/shared' if Environment.hasSpark and not self.forcePandas and os.path.exists('/home/spark/shared') else tempfile.gettempdir()
            with tempfile.NamedTemporaryFile(delete=False, dir=tdir) as f:
                bytesDownloaded = self.write(urlopen(req), f)
                path = f.name   
            if url.endswith(".zip") or zipfile.is_zipfile(path):
                #unzip first and get the first file in it
                print("Extracting first item in zip file...")
                import shutil
                zfile = zipfile.ZipFile(path, 'r')
                if len(zfile.filelist)==0:
                    raise(Exception("Error: zip file is empty"))
                with tempfile.NamedTemporaryFile(delete=False, dir=tdir) as zf:
                    with zfile.open( zfile.filelist[0], 'r') as first_file:
                        print("File extracted: {}".format(first_file.name))
                        shutil.copyfileobj( first_file, zf)
                    path = zf.name
            elif url.endswith(".gz"):
                import gzip
                import shutil
                with tempfile.NamedTemporaryFile(delete=False, dir=tdir) as zf:
                    with gzip.open(path, 'rb') as first_file:
                        print("File extracted: {}".format(first_file.name))
                        shutil.copyfileobj( first_file, zf)
                    path = zf.name

            self.dataDef["path"] = path
            self.dataDef["transient"] = True
            global dataDefs
            dataDefs[url] = self.dataDef
        if path:
            try:
                if bytesDownloaded > 0:
                   print("Downloaded {} bytes".format(bytesDownloaded))
                print("Creating {1} DataFrame for '{0}'. Please wait...".\
                    format(displayName, 'pySpark' if Environment.hasSpark and not self.forcePandas else 'pandas'))
                return dataLoader(path, self.dataDef.get("schema", None))
            finally:
                print("Successfully created {1} DataFrame for '{0}'".\
                    format(displayName, 'pySpark' if Environment.hasSpark and not self.forcePandas else 'pandas'))

    def report(self, bytes_so_far, chunk_size, total_size):
        if useProgressMonitor:
            if bytes_so_far == 0:
                display( HTML( """
                    <div>
                        <span id="pm_label{0}">Starting download...</span>
                        <progress id="pm_progress{0}" max="100" value="0" style="width:200px"></progress>
                    </div>""".format(self.prefix)
                    )
                )
            else:
                percent = float(bytes_so_far) / total_size
                percent = round(percent*100, 2)
                display(
                    Javascript("""
                        $("#pm_label{prefix}").text("{label}");
                        $("#pm_progress{prefix}").attr("value", {percent});
                    """.format(prefix=self.prefix, label="Downloaded {0} of {1} bytes".format(bytes_so_far, total_size), percent=percent))
                )

    def write(self, response, file, chunk_size=8192):
        total_size = response.headers['Content-Length'].strip() if 'Content-Length' in response.headers else 100
        total_size = int(total_size)
        bytes_so_far = 0

        self.report(bytes_so_far, chunk_size, total_size)

        while 1:
            chunk = response.read(chunk_size)
            bytes_so_far += len(chunk)
            if not chunk:
                break
            file.write(chunk)
            total_size = bytes_so_far if bytes_so_far > total_size else total_size
            self.report(bytes_so_far, chunk_size, total_size)

        return bytes_so_far



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from six import iteritems
from .storage import *

USER_PREFERENCES_TBL_NAME = "USER_PREFERENCES"

class __UserPreferences(Storage):
    def __init__(self):
        self._initTable(USER_PREFERENCES_TBL_NAME,
        '''
            PREF_KEY     TEXT  NOT NULL,
            PREF_VALUE   TEXT  NOT NULL,
            PRIMARY KEY (PREF_KEY)
        ''')

userPrefStorage = __UserPreferences()

def getUserPreference(prefKey, defaultValue=None):
    retValue = userPrefStorage.fetchOne("""
            SELECT * from {0} WHERE PREF_KEY='{1}'
        """.format(USER_PREFERENCES_TBL_NAME,prefKey), 
        lambda row: row["PREF_VALUE"]
    )
    return defaultValue if retValue is None else retValue

def setUserPreference(prefKey, prefValue):
    if getUserPreference(prefKey) is not None:
        userPrefStorage.update("""
            UPDATE {0} SET PREF_VALUE='{2}' WHERE PREF_KEY = '{1}'
        """.format(USER_PREFERENCES_TBL_NAME,prefKey, prefValue))
    else:
        userPrefStorage.insert("""
            INSERT INTO {0} (PREF_KEY, PREF_VALUE)
            VALUES ('{1}','{2}')
        """.format(USER_PREFERENCES_TBL_NAME, prefKey, prefValue))

class user_preferences:
    """
    Decorator for classed to automatically add fields getter and setter that are persisted
    in the pixiedust database
    example usage:
    @user_preferences(key1="value1", key2="value2")
    class Test():
        pass

    t = Test()
    print(t.key1)
    t.key1 = "value11"
    print(t.key1)
    """
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __call__(self, cls):
        for key, default_value in iteritems(self.kwargs):
            def create_property_fn(key, default_value):
                prop_name = "_" + key
                preference_key_name = "{}_{}".format(cls.__name__, key)
                def prop_get(self):
                    if not hasattr(self, prop_name):
                        setattr(self, prop_name, getUserPreference(preference_key_name, default_value))
                    return getattr(self, prop_name)
                def prop_set(self, value):
                    setUserPreference(preference_key_name, value)
                    setattr(self, prop_name, value)
                setattr(cls, key, property(prop_get, prop_set))
            create_property_fn(key, default_value)
        return cls



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from pixiedust.display import display
from pixiedust.display.display import *
from pixiedust.utils.shellAccess import ShellAccess
from pixiedust.utils import Logger
from six import iteritems, string_types
from collections import OrderedDict, namedtuple
import base64
import inspect
import sys
from functools import partial
from jinja2 import Template
from IPython.utils.io import capture_output

def route(**kw):
    def route_dec(fn):
        fn.pixiedust_route = kw
        if hasattr(fn, "fn"):
            fn.fn.persist_args = kw.pop("persist_args", None)
        return fn
    return route_dec

@Logger()
class captureOutput(object):
    """
    Decorator used for routes that allows using external libraries for generating
    the html fragment. 
    When using this decorator the route doesn't need to return a string. If it does
    it will be ignored.
    Must be declared in after the route decorator. 
    captureOutput and templateArgs should not be used together
        from pixiedust.display.app import *
        import matplotlib.pyplot as plt
        import numpy as np
        @PixieApp
        class Test():
            @route()
            @captureOutput
            def mainScreen(self):
                t = np.arange(0.0, 2.0, 0.01)
                s = 1 + np.sin(2*np.pi*t)
                plt.plot(t, s)

                plt.xlabel('time (s)')
                plt.ylabel('voltage (mV)')
                plt.title('About as simple as it gets, folks')
                plt.grid(True)
                plt.savefig("test.png")
                plt.show()
        Test().run()
    
    """
    def __init__(self, fn):
        self.fn = fn

    def convert_html(self, output):
        if "text/html" in output.data:
            return output._repr_html_()
        elif "image/png" in output.data:
            return """<img alt="image" src="data:image/png;base64,{}"><img>""".format(
                base64.b64encode(output._repr_png_()).decode("ascii")
            )
        elif "application/javascript" in output.data:
            return """<script type="text/javascript">{}</script>""".format(output._repr_javascript_())
        elif "text/markdown" in output.data:
            import markdown
            return markdown.markdown(output._repr_mime_("text/markdown"))
        self.debug("Unused output: {}".format(output.data.keys()))
        return ""

    def __get__(self, instance, instance_type):
        wrapper_fn = partial(self.wrapper, instance)
        wrapper_fn.org_fn = self.fn
        return wrapper_fn

    def wrapper(self, instance, *args, **kwargs):
        ret_html = None
        with capture_output() as buf:
            ret_html = self.fn(instance, *args, **kwargs)
        captured_output = "\n".join([self.convert_html(output) for output in buf.outputs])
        if ret_html is not None:
            captured_output += "\n" + ret_html
        return captured_output

class templateArgs(object):
    """
    Decorator that enables using local variable in a Jinja template.
    Must be used in conjunction with route decorator and declared after
        from pixiedust.display.app import *
        @PixieApp
        class Test():
            @route()
            @templateArgs
            def mainScreen(self):
                var1 = 'something computed'
                return "<div>Accessing local variable {{var1}} from a jinja template"
        Test().run()
    """
    TemplateRetValue = namedtuple('TemplateRetValue', ['ret_value', 'locals'])
    def __init__(self, fn):
        self.fn = fn

    def __get__(self, instance, instance_type):
        wrapper_fn = partial(self.wrapper, instance)
        wrapper_fn.org_fn = self.fn
        return wrapper_fn

    def wrapper(self, instance, *args, **kwargs):
        locals = [{}]
        def tracer(frame, event, arg):
            if event == "return":
                locals[0] = frame.f_locals.copy()
                if 'self' in locals[0]:
                    del locals[0]['self']
        sys.setprofile(tracer)
        try:
            ret_value = self.fn(instance, *args, **kwargs)
            return templateArgs.TemplateRetValue(ret_value, locals[0])
        finally:
            sys.setprofile(None)

#Global object enables system wide customization of PixieApp run option
pixieAppRunCustomizer = None

def runPixieApp(app, parent_pixieapp=None, entity=None, **kwargs):
    options = kwargs.get("options", {})
    if options.pop("new_parent_prefix", "true") == "true":
        options.pop("prefix", None)  #child pixieapp should have its own prefix
    if isinstance(app, PixieDustApp):
        app.run(entity, **kwargs)
    elif isinstance(app, string_types):
        parts = app.split('.')
        instance_app = None
        if len(parts) > 1:
            instance_app = getattr(__import__('.'.join(parts[:-1]), None, None, [parts[-1]], 0), parts[-1])()
        else:
            instance_app = ShellAccess[parts[-1]]()
        if parent_pixieapp is not None:
            instance_app.parent_pixieapp = parent_pixieapp if isinstance(parent_pixieapp, PixieDustApp) else ShellAccess[parent_pixieapp]
            instance_app.parent_pixieapp.add_child(instance_app)
        kwargs["is_running_child_pixieapp"] = True
        instance_app.run(entity, **kwargs)
    else:
        raise ValueError("Invalid argument to runPixieApp. Only PixieApp or String allowed")

class RunInPixieDebugger(Exception):
    "Marker exception to start the PixieDebugger"
    pass

@Logger()
class PixieDustApp(Display):

    routesByClass = {}

    def __init__(self, options=None, entity=None, dataHandler=None):
        super(PixieDustApp, self).__init__(options, entity, dataHandler)
        if not hasattr(self, "parent_pixieapp"):
            self.parent_pixieapp = None
        self.exceptions = {}
        if not hasattr(self, "metadata"):
            self.metadata = None
            self.empty_metadata = False

    def append_metadata(self, value):
        if self.empty_metadata:
            self.metadata = {}
            self.empty_metadata = False
        else:
            self.metadata = self.metadata or {}
        self.metadata.update(value)

    def getOptionValue(self, optionName):
        option = None
        if self.metadata:
            option = self.metadata.get(optionName, None)
        if option is None:
            #check if the key is an field of the class
            option = getattr(self.entity, optionName) if self.entity is not None and hasattr(self.entity, optionName) else None
            #make sure we don't have a conflict with an existing function
            if callable(option):
                option = None
        if option is None:
            option = self.options.get(optionName, None)
        return option

    def matchRoute(self, route):
        for key, value in iteritems(route):
            option = self.getOptionValue(key)
            if  (option is None and value == "*") or (value != "*" and option != value):
                return False
        return True

    def has_persist_args(self, method):
        if isinstance(method, partial) and hasattr(method, "org_fn"):
            method = method.org_fn
        return getattr(method, "persist_args", None) is not None

    def injectArgs(self, method, **kwargs):
        if isinstance(method, partial) and hasattr(method, "org_fn"):
            method = method.org_fn
        argspec = inspect.getargspec(method)
        args = argspec.args
        if len(args) > 0:
            args = args[1:] if hasattr(method, "__self__") or args[0] == 'self' else args
        return OrderedDict(zip([a for a in args], [self.getOptionValue(arg) or kwargs.get(arg, None) for arg in args]))

    def invoke_route(self, class_method, **kwargs):
        "Programmatically invoke a route from arguments"
        try:
            injectedArgs = self.injectArgs(class_method, **kwargs)
            retValue = class_method(*list(injectedArgs.values()))
        finally:
            if isinstance(retValue, templateArgs.TemplateRetValue):
                injectedArgs.update(retValue.locals)
                retValue = retValue.ret_value
            if isinstance(retValue, string_types):
                retValue = self.renderTemplateString(retValue, **injectedArgs)
        return retValue

    def __getattr__(self, name):
        if ShellAccess[name] is not None:
            return ShellAccess[name]
        if name != "__pd_gateway_namespace__" and hasattr(self, "__pd_gateway_namespace__"):
            name = self.__pd_gateway_namespace__ + name
            if ShellAccess[name] is not None:
                return ShellAccess[name]
        raise AttributeError("{} attribute not found".format(name))

    def hook_msg(self, msg):
        msg['content']['metadata']['pixieapp_metadata'] = self.metadata
        self.empty_metadata = True
        return msg

    def render(self):
        from IPython.core.interactiveshell import InteractiveShell
        display_pub = InteractiveShell.instance().display_pub
        try:
            display_pub.register_hook(self.hook_msg)      
            super(PixieDustApp, self).render()
        except RunInPixieDebugger:
            self.options.pop("handlerId")
            self.options['new_parent_prefix'] = False
            entity={
                "breakpoints": ["{}.{}".format(self.__pixieapp_class_name__, self.breakpoints)],
                "code": self.get_pd_controls()["command"]
            }
            self.breakpoints = None
            runPixieApp(
                "pixiedust.apps.debugger.PixieDebugger", 
                parent_pixieapp=self, 
                entity=entity,
                **{"options":self.options})
        finally:
            display_pub.unregister_hook(self.hook_msg)

    def doRender(self, handlerId):
        if self.__class__.__name__ in PixieDustApp.routesByClass:
            defRoute = None
            retValue = None
            injectedArgs = {}
            try:
                dispatchKey = "widgets" if "widget" in self.options else "routes"
                for t in PixieDustApp.routesByClass[self.__class__.__name__][dispatchKey]:
                    if not t[0]:
                        defRoute = t[1]
                    elif self.matchRoute(t[0]):
                        if self.breakpoints == t[1]:
                            self.debug("Breakpoint matched, Invoking PixieDebugger")
                            raise RunInPixieDebugger()
                        self.debug("match found: {}".format(t[0]))
                        meth = getattr(self, t[1])
                        injectedArgs = self.injectArgs(meth)
                        self.debug("Injected args: {}".format(injectedArgs))
                        if self.metadata is None and self.has_persist_args(meth):
                            self.metadata = {key:self.getOptionValue(key) for key,_ in iteritems(t[0])}
                        try:
                            self.exceptions.pop(t[1], None)
                            retValue = meth(*list(injectedArgs.values()))
                        except:
                            self.exceptions[t[1]] = self.get_pd_controls()["command"]
                            raise
                        return
                if defRoute:
                    if self.breakpoints == defRoute:
                        self.debug("Breakpoint matched, Invoking PixieDebugger")
                        raise RunInPixieDebugger()
                    try:
                        self.exceptions.pop(defRoute, None)
                        retValue = getattr(self, defRoute)()
                    except:
                        self.exceptions[defRoute] = self.get_pd_controls()["command"]
                        raise
                    return
            finally:
                if isinstance(retValue, templateArgs.TemplateRetValue):
                    injectedArgs.update(retValue.locals)
                    retValue = retValue.ret_value
                if isinstance(retValue, string_types):
                    if self.getBooleanOption("nostore_isrunningchildpixieapp", False):
                        self.options.pop("nostore_isrunningchildpixieapp", None)
                        retValue = """<div id="wrapperHTML{{prefix}}" pixiedust="{{this.get_pd_controls(black_list=['nostore_isrunningchildpixieapp'])|tojson|htmlAttribute}}">""" + retValue + """</div>"""
                    self._addHTMLTemplateString(retValue, **injectedArgs)
                elif isinstance(retValue, Template):
                    self._addHTML(
                        retValue.render(self._getTemplateArgs(
                                resModule=".".join(type(self).__name__.split(".")[:-1]), 
                                **injectedArgs
                            )
                        )
                    )
                elif isinstance(retValue, dict):
                    body = self.renderTemplateString(retValue.get("body", ""))
                    jsOnLoad = self.renderTemplateString(retValue.get("jsOnLoad", ""))
                    jsOK = self.renderTemplateString(retValue.get("jsOK", ""))
                    dialogRoot = retValue.get("dialogRoot", None)
                    if dialogRoot is not None:
                        jsOnLoad = """pixiedust.dialogRoot="{}";\n{}""".format(self.renderTemplateString(dialogRoot), jsOnLoad)
                    if body is not None:
                        self._addHTMLTemplateString("""
                        {{body}}
                        <pd_dialog>
                            <pd_onload>{{jsOnLoad|htmlAttribute}}</pd_onload>
                            <pd_ok>{{jsOK|htmlAttribute}}</pd_ok>
                        </pd_dialog>
                        """, body=body, jsOnLoad=jsOnLoad, jsOK=jsOK)

        print("Didn't find any routes for {}. Did you forget to define a default route?".format(self))

    pixieapp_child_prefix = "__pixieapp_child__"
    @property
    def pixieapp_children(self):
        return {var:getattr(self, var) for var in dir(self) if var.startswith(PixieDustApp.pixieapp_child_prefix)}

    def add_child(self, instance_app):
        var_name = "{}{}".format(
            PixieDustApp.pixieapp_child_prefix,
            len([var for var in dir(self) if var.startswith(PixieDustApp.pixieapp_child_prefix)])
        )
        setattr(self, var_name, instance_app)

    def get_custom_options(self):
        return {}

    def getDialogOptions(self):
        return {}

@Logger()
def PixieApp(cls):
    #reset the class routing in case the cell is being run multiple time
    clsName = "{}_{}_Display".format(inspect.getmodule(cls).__name__, cls.__name__)
    PixieDustApp.routesByClass[clsName] = {"routes":[], "widgets":[]}
    #put the routes that define a widget in a separate bucket

    def walk(cl):
        for name, method in iteritems(cl.__dict__):
            if hasattr(method, "pixiedust_route"):
                if "widget" in method.pixiedust_route:
                    PixieDustApp.routesByClass[clsName]["widgets"].append( (method.pixiedust_route,name) )
                else:
                    PixieDustApp.routesByClass[clsName]["routes"].append( (method.pixiedust_route,name) )
        for c in [c for c in cl.__bases__]:
            walk(c)
    walk(cls)

    #re-order the routes according to the number of constraints e.g. from more to less specific
    p = PixieDustApp.routesByClass[clsName]["routes"]
    PixieDustApp.routesByClass[clsName]["routes"] = [p[a[1]] for a in sorted([(len(a[0]), i) for i,a in enumerate(p)], reverse=True)]

    def __init__(self, options=None, entity=None, dataHandler=None):
        PixieDustApp.__init__(self, options or {}, entity, dataHandler)

    def getPixieAppEntity(self):
        return self.pixieapp_entity if hasattr(self, "pixieapp_entity") else None

    def formatOptions(self,options):
        """Helper method that convert pd options from Json format to pixieApp html attribute compliant format"""
        return ';'.join(["{}={}".format(key,value) for (key, value) in iteritems(options)])

    def decoName(cls, suffix):
        return "{}_{}_{}".format(cls.__module__, cls.__name__, suffix)

    def run_method_with_super_classes(cls, instance, method_name):
        fctSet = set()
        for cl in reversed(inspect.getmro(cls)):
            if hasattr(cl, 'setup'):
                f = getattr(cl, 'setup')
                if f not in fctSet and callable(f):
                    fctSet.add(f)
                    f(instance)

    def run(self, entity=None, **kwargs):
        self.breakpoints = kwargs.pop("debug_route", "")
        is_running_child_pixieapp = kwargs.pop("is_running_child_pixieapp", False)
        for key, value in iteritems(kwargs):
            setattr(self, key, value)
        self.pixieapp_entity = entity
        var = None
        if self.parent_pixieapp is not None:
            def find_child_var(parent, child):
                for c in dir(parent):
                    if getattr(parent, c) == child:
                        return c
            def find_notebook_var(parent):
                for key in ShellAccess.keys():
                    notebook_var = ShellAccess[key]
                    if notebook_var is parent and key != "self" and not key.startswith("_"):
                        return key, notebook_var
                # we didn't find it, maybe it's nested
                if parent.parent_pixieapp is not None:
                    key, notebook_var = find_notebook_var(parent.parent_pixieapp)
                    if key is not None:
                        # find the current object by name and append it to the key         
                        return (key + "." + find_child_var(parent.parent_pixieapp, parent), notebook_var)
                return None,None
            parent_key, notebook_var = find_notebook_var(self.parent_pixieapp)
            for child_key, child in iteritems(notebook_var.pixieapp_children):
                if child is self:
                    var = "{}.{}".format(parent_key, child_key)
                    break
        else:
            for key in ShellAccess.keys():
                notebook_var = ShellAccess[key]
                if notebook_var is self:
                    var = key
                    break

        if not hasattr(self, "pd_initialized"):
            run_method_with_super_classes(cls, self, "setup")
            self.nostore_params = True
            self.pd_initialized = True

        instance_namespace = ""
        if is_running_child_pixieapp:
            cell_id = kwargs.get("options", {}).get("cell_id", None)
            if cell_id:
                instance_namespace = "_" + cell_id
        if not var:
            #If we're here, the user must have created the instance inline, assign a variable dynamically
            var = cls.__name__ + "_instance" + instance_namespace
            ShellAccess[var] = self

        self.runInDialog = kwargs.get("runInDialog", "false") is "true"
        options = {
            "nostore_pixieapp": var,
            "nostore_ispix":"true",
            "runInDialog": "true" if self.runInDialog else "false"
        }
        if is_running_child_pixieapp:
            options["nostore_isrunningchildpixieapp"] = "true"
    
        #update with any custom options that the pixieapp may have
        options.update(self.get_custom_options())

        if self.runInDialog:
            options.update(self.getDialogOptions())

        options.update({'handlerId': decoName(cls, "id")})
        if "options" in kwargs and isinstance(kwargs['options'], dict):
            options.update(kwargs['options'])
        if pixieAppRunCustomizer is not None and callable(getattr(pixieAppRunCustomizer, "customizeOptions", None)):
            pixieAppRunCustomizer.customizeOptions(options)

        opts = [(k, str(v).lower() if isinstance(v, bool) else v) for (k,v) in iteritems(options) if v is not None]
        display_call = "display" if ShellAccess['display'] is not None and ShellAccess['display'].__module__ == "pixiedust.display" else "pixiedust.display"
        s = "{}({}{})".format(display_call, var, reduce(lambda k,v: k + "," + v[0] + "='" + str(v[1]) + "'", opts, ""))

        try:
            sys.modules['pixiedust.display'].pixiedust_display_callerText = s
            self._app_starting = True   #App lifecycle flag
            parts = var.split(".")
            locals()[parts[0]] = ShellAccess[parts[0]]
            self.debug("Running with command: {} and var {}".format(s, var))
            return eval(s, globals(), locals())
        finally:
            self._app_starting = False
            del sys.modules['pixiedust.display'].pixiedust_display_callerText
        
    displayClass = type( decoName(cls, "Display"), (cls,PixieDustApp, ),{
        "__init__": __init__, 
        "run": run, 
        "getPixieAppEntity":getPixieAppEntity,
        "__pixieapp_class_name__": cls.__name__
    })
    ShellAccess["newDisplayClass"] = displayClass

    def prettyFormat(o):
        return "{} at {}".format(o, id(o))
    
    @addId
    def getMenuInfo(self, entity, dataHandler=None):
        if entity is displayClass or entity.__class__ is displayClass:
            return [{"id": decoName(cls, "id")}]
        return []

    def newDisplayHandler(self, options, entity):
        if entity is displayClass or entity.__class__ is displayClass:
            entity.__init__(options, entity)
            return entity
        elif options.get("nostore_pixieapp") is not None:
            from pixiedust.utils.shellAccess import ShellAccess
            papp = ShellAccess[options.get("nostore_pixieapp")]
            if papp is not None and hasattr(papp, "newDisplayHandler"):
                fn = papp.newDisplayHandler
                if callable(fn):
                    return fn(options, entity)
        return None

    displayHandlerMetaClass = type( decoName(cls, "Meta"), (DisplayHandlerMeta,), {
            "getMenuInfo": getMenuInfo,
            "newDisplayHandler": newDisplayHandler
        })

    displayHandlerMeta = displayHandlerMetaClass()
    ShellAccess["displayHandlerMeta"] = displayHandlerMeta
    registerDisplayHandler( displayHandlerMeta )
    return displayClass



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.app import *
from pixiedust.display.streaming.data import *
from pixiedust.display.streaming.bokeh import *
from pixiedust.utils import Logger
import requests

useConfluent = False

@PixieApp
@Logger()
class MessageHubStreamingApp():
    def setup(self):
        self.streamingDisplay = None
        self.streamingData = None
        self.contents = []
        self.schemaX = None
        self.capturing = False
        self.captureVarName = "messageHubData"

    def newDisplayHandler(self, options, entity):
        if self.streamingDisplay is None:
            self.streamingDisplay = LineChartStreamingDisplay(options, entity)
        else:
            self.streamingDisplay.options = options
        return self.streamingDisplay
    
    def getTopics(self):
        rest_endpoint = "https://kafka-rest-prod01.messagehub.services.us-south.bluemix.net:443"
        headers = {
            'X-Auth-Token': self.credentials["api_key"],
            'Content-Type': 'application/json'
        }        
        response = requests.get('{}/topics'.format(rest_endpoint),headers = headers)
        if response.status_code != requests.codes.ok:
            response.raise_for_status()
        return response.json()
    
    @route()
    def mainScreen(self):
        return """
<div class="well" style="text-align:center">
    <div style="font-size:x-large">MessageHub Streaming Browser.</div>
    <div style="font-size:large">Click on a topic to start</div>
</div>

{%for topic in this.getTopics()%}
    {%if loop.first or ((loop.index % 4) == 1)%}
<div class="row">
    <div class="col-sm-2"/>
    {%endif%}
    <div pd_options="topic=$val(topic{{loop.index}}{{prefix}})" class="col-sm-2" style="border: 1px solid lightblue;margin: 10px;border-radius: 25px;cursor:pointer;
        min-height: 150px;background-color:#e6eeff;display: flex;align-items: center;justify-content:center">
        <span id="topic{{loop.index}}{{prefix}}">{{topic}}</span>
    </div>
    {%if loop.last or ((loop.index % 4) == 0)%}
    <div class="col-sm-2"/>
</div>
    {%endif%}
{%endfor%}
        """
    
    def displayNextTopics(self):
        payload = self.streamingData.getNextData()
        if payload is not None and len(payload)>0:
            self.contents = self.contents + payload
            self.contents = self.contents[-10:]                
            html = ""
            for event in self.contents:
                html += "{}<br/>".format(json.dumps(event))
            print(html)
            
    @route(topic="*",streampreview="*",schemaX="*")
    def showChart(self, schemaX):
        self.schemaX = schemaX
        self.avgChannelData = self.streamingData.getStreamingChannel(self.computeAverages)
        return """
<div class="well" style="text-align:center">
    <div style="font-size:x-large">Real-time chart for {{this.schemaX}}(average).</div>
</div>
<style>
.bk-root{
display:flex;
justify-content:center;
}
</style>
<div pd_refresh_rate="1000" pd_entity="avgChannelData"></div>
        """
    
    def computeAverages(self, avg, newData):
        newValue = []
        for jsonValue in newData:
            if self.schemaX in jsonValue:
                thisValue = float(jsonValue[self.schemaX])
                avg = thisValue if avg is None else (avg + thisValue)/2
                newValue.append(avg)
        return newValue, avg
    
    def captureDataInNotebook(self, acc, newData):
        if self.capturing:
            for jsonValue in newData:
                ShellAccess[self.captureVarName].append(jsonValue)
        return None,None
    
    def toggleCapture(self):
        self.capturing = not self.capturing
        if self.capturing:
            capturedData = ShellAccess[self.captureVarName]
            if capturedData is None:
                ShellAccess[self.captureVarName] = []
                capturedData = ShellAccess[self.captureVarName]
            
        print("Capturing in var messageHubData... Click to stop" if self.capturing else "Capture Data")        
    
    @route(topic="*",streampreview="*")
    def createStreamWidget(self, streampreview):
        if streampreview=="realtimeChart":
            return """
<div>
    {%for key in this.streamingData.schema%}
    {%if loop.first%}
    <div class="well" style="text-align:center">
        <div style="font-size:x-large">Create a real-time chart by selecting a field.</div>
    </div>
    {%endif%}
    <div class="radio" style="margin-left:20%">
      <label>
          <input type="radio" pd_options="streampreview=""" + streampreview + """;schemaX=$val(schemaX{{loop.index}}{{prefix}})" 
              id="schemaX{{loop.index}}{{prefix}}" pd_target="realtimeChartStreaming{{prefix}}" 
              name="schemaX" value="{{key}}">{{key}}
      </label>
    </div>
    {%endfor%}
</div>"""
        return """<div pd_refresh_rate="1000" pd_script="self.displayNextTopics()"></div>"""
        
    @route(topic="*")
    def previewTopic(self, topic):
        self.topic = topic
        if self.streamingData is not None:
            self.streamingData.close()
        Adapter = MessagehubStreamingAdapter
        if useConfluent:
            from pixiedust.display.streaming.data.messageHubConfluent import MessagehubStreamingAdapterConfluent
            Adapter = MessagehubStreamingAdapterConfluent
            self.debug("Using Confluent Kafka")
        self.streamingData = Adapter( self.topic, self.credentials["username"], self.credentials["password"], self.credentials.get("prod", True) )
        self.streamingData.getStreamingChannel(self.captureDataInNotebook)
        return """
<div class="well" style="text-align:center">
    <div style="font-size:x-large">Preview raw JSON Data from MessageHub.</div>
    <div id="tools{{prefix}}" style="text-align:center">
        <button type="button" class="btn btn-default" pd_target="captureBtn{{prefix}}">
            <span class="no_loading_msg" id="captureBtn{{prefix}}">Capture Data</span>
            <pd_script>self.toggleCapture()</pd_script>
        </button>
    </div>
</div>
<div class="row">
    <div class="col-sm-12" id="targetstreaming{{prefix}}">
        <div pd_refresh_rate="1000" style="white-space:nowrap;overflow-x:auto;border:aliceblue 2px solid;height:17em;line-height:1.5em">
            <pd_script>self.displayNextTopics()</pd_script>
            <div style="width:100px;height:60px;left:47%;position:relative">
                <i class="fa fa-circle-o-notch fa-spin" style="font-size:48px"></i>
            </div>
            <div style="text-align:center">Waiting for data from MessageHub</div>
        </div>
    </div>
</div>
<div class="row" id="realtimeChartStreaming{{prefix}}">
    <div pd_refresh_rate="4000" pd_options="streampreview=realtimeChart">
    </div>
</div>
        """


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
import warnings
import inspect
import sys
import argparse
from pixiedust.display.app import *
from pixiedust.utils import Logger
from pixiedust.utils.astParse import get_matches_lineno
from six import iteritems, PY3
from IPython.core.magic import (Magics, magics_class, line_cell_magic)
from IPython.core.getipython import get_ipython

@magics_class
class PixiedustDebuggerMagics(Magics):
    def __init__(self, shell):
        super(PixiedustDebuggerMagics,self).__init__(shell=shell)

    @line_cell_magic
    def pixie_debugger(self, line, cell=None):
        args = None
        if line is not None:
            parser = argparse.ArgumentParser(description='PixieDebugger options')
            parser.add_argument('-b', '--breakpoints', nargs='*', help='set breakpoints')
            args = parser.parse_args(line.split())
        PixieDebugger().run({
            "breakpoints": args.breakpoints if args is not None else [],
            "code": cell
        })

try:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        get_ipython().register_magics(PixiedustDebuggerMagics)
except NameError:
    #IPython not available we must be in a spark executor
    pass

class NoTraceback(Exception):
    "Marker exception for signaling that no traceback was found"
    pass

@PixieApp
@Logger()
class PixieDebugger():
    def initialize(self):
        self.is_post_mortem = False
        self.code =  ""
        self.breakpoints = []
        self.is_post_mortem = self.pixieapp_entity is None or self.pixieapp_entity["code"] is None
        if self.is_post_mortem:
            if not hasattr(sys, "last_traceback") or sys.last_traceback is None:
                raise NoTraceback()
            if self.options.get("debug_route", "false" ) == "true":
                stack = [tb.function if PY3 else tb[3] for tb in inspect.getinnerframes(sys.last_traceback)]
                method_name = stack[-1]
                for method in reversed(stack):
                    code = self.parent_pixieapp.exceptions.get(method, None)
                    if code:
                        method_name = method
                        break
                self.pixieapp_entity = {
                    "breakpoints": ["{}.{}".format(self.parent_pixieapp.__pixieapp_class_name__, method_name)],
                    "code": code or ""
                }
                self.debug("Invoking PixieDebugger for route with {}".format(self.pixieapp_entity))
                self.is_post_mortem = False

        if not self.is_post_mortem:
            self.code = self.renderTemplateString("""
def pixie_run():
    import pdb
    pdb.set_trace()
{{this.pixieapp_entity["code"]|indent(4, indentfirst=True)}}
pixie_run()
        """.strip())
            bps_lines = set()
            bps_fns = set()
            for breakpoint in self.pixieapp_entity["breakpoints"] or []:
                try:
                    bps_lines.add( int(breakpoint))
                except ValueError:
                    matches = get_matches_lineno(self.code, breakpoint)
                    if len(matches) == 0:
                        bps_fns.add(breakpoint)
                    else:
                        for match in matches:
                            bps_lines.add(match)
            self.breakpoints = list(sorted(bps_lines)) + list(sorted(bps_fns))
        else:
            self.code="""
import pdb
pdb.pm()
            """

    @property
    def pm_prefix(self):
        return "$$" if self.is_post_mortem else ""

    @route()
    def main_screen(self):
        try:
            self.initialize()
        except NoTraceback:
            return "<div>Nothing to debug because no Traceback has been produced</div>"
        return self.env.getTemplate("mainScreen.html")

    @route(done="*")
    def done_screen(self):
        return "<div>Thank you for using PixieDust Debugger</div>"



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.utils import Logger
from .brunelBaseDisplay import BrunelBaseDisplay

@PixiedustRenderer(id="barChart")
@Logger()
class BarChartRenderer(BrunelBaseDisplay):
    def isSubplot(self):
        return self.options.get("charttype", "grouped") == "subplots"

    def getExtraFields(self):
        if not self.isSubplot() and len(self.getValueFields()) > 1:
            #no clusterby if we are grouped and multiValueFields
            return []

        clusterby = self.options.get("clusterby")
        return [clusterby] if clusterby is not None else []

    def compute_brunel_magic(self):
        parts = ["bar"]
        if self.options.get("orientation", "vertical") == "horizontal":
            parts.append("transpose")

        clusterby = list(filter(None, [self.options.get("clusterby", "")]))

        if self.options.get("charttype", "grouped") == "stacked":
            parts.append("stack")
            parts.append("x({})".format(",".join(self.getKeyFields())))
        else:
            parts.append("x({})".format(",".join(self.getKeyFields() + self.getValueFields())))
        parts.append("y({})".format(",".join(self.getValueFields())))
        parts.append(self.get_sort())
        parts.append("color({})".format(clusterby[0] if len(clusterby) > 0 else self.getKeyFields()[0]))
        return parts



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.streaming import *
from six import string_types, iteritems
import json
import os
try:
    from confluent_kafka import Consumer
except:
    from confluent_kafka_prebuilt import Consumer

class MessagehubStreamingAdapterConfluent(StreamingDataAdapter):
    def __init__(self, topic, username, password, prod=True):
        caLocation = '/etc/ssl/cert.pem'
        if not os.path.exists(caLocation):
            caLocation = '/etc/pki/tls/cert.pem'
        conf = {
            'client.id': 'pixieapp.client.id',
            'group.id': 'pixieapp.group',
            'security.protocol': 'SASL_SSL',
            'sasl.mechanisms': 'PLAIN',
            'ssl.ca.location': caLocation,
            "bootstrap.servers": ','.join(["kafka0{}-{}.messagehub.services.us-south.bluemix.net:9093".format(i, "prod01" if prod else "stage1") for i in range(1,6)]),
            "sasl.username": username,
            "sasl.password": password,
            'api.version.request': True
        }
        self.consumer = Consumer(conf)
        self.consumer.subscribe([topic])
        self.schema = {}
        self.sampleDocCount = 0
        
    def close(self):
        self.consumer.unsubscribe()
        self.consumer.close() 
        
    def tryCast(self, value, t):
        try:
            return t(value)
        except:
            return None
        
    def inferType(self, value):
        if isinstance(value, string_types):
            value = self.tryCast(value, int) or self.tryCast(value, long) or self.tryCast(value, float) or value
        return "integer" if value.__class__==int else "float" if value.__class__ == float else "string"
        
    def inferSchema(self, eventJSON):
        if self.sampleDocCount > 20:
            return
        for key,value in iteritems(eventJSON):
            if not key in self.schema:
                self.schema[key] = self.inferType(value)
        self.sampleDocCount = self.sampleDocCount + 1 
    
    def doGetNextData(self):
        msgs = []
        msg = self.consumer.poll(1)
        if msg is not None and msg.error() is None:
            jsonValue = json.loads(msg.value())
            self.inferSchema(json.loads(msg.value()))
            msgs.append(jsonValue)
        return msgs
    
    def close(self):
        self.consumer.close()


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from abc import abstractmethod, ABCMeta
from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.display.chart.renderers.baseChartDisplay import commonChartOptions
from pixiedust.utils import Logger
from pixiedust.utils.shellAccess import ShellAccess
from six import with_metaclass
from IPython.display import display as ipythonDisplay, HTML
from IPython.utils.io import capture_output
from IPython.core.getipython import get_ipython
from ..baseChartDisplay import BaseChartDisplay

@PixiedustRenderer(rendererId="brunel")
@Logger()
class BrunelBaseDisplay(with_metaclass(ABCMeta, BaseChartDisplay)):
    def convert_html(self, output):
        if "text/html" in output.data:
            return output._repr_html_()
        elif "application/javascript" in output.data:
            return """<script type="text/javascript">{}</script>""".format(output._repr_javascript_())
        self.debug("Unused output: {}".format(output.data.keys()))
        return ""

    @abstractmethod
    def compute_brunel_magic(self):
        pass

    def use_online_js(self):
        return self.getBooleanOption("chart_share", False) or self.is_gateway

    def complete_magic(self, magic_parts):
        dynamicfilter = self.options.get("dynamicfilter")
        if dynamicfilter is not None:
            magic_parts.append("filter({})".format(dynamicfilter))
        magic_parts.append(":: width={}, height={}".format(
            int(self.getPreferredOutputWidth()), int(self.getPreferredOutputHeight())
        ))
        if self.use_online_js():
            magic_parts.append(", online_js=True")
        return " ".join(magic_parts)

    @commonChartOptions
    def getChartOptions(self):
        return [
            {
                'name': 'dynamicfilter',
                'description': 'Dynamic Filter',
                'refresh': False,
                'metadata': {
                    'type': 'dropdown',
                    'values': ['None'] + sorted(self.getKeyFields() + self.getValueFields()),
                    'default': ''
                },
                'validate': lambda option:\
                    (option in self.getFieldNames() and (option in self.getKeyFields() or option in self.getValueFields()),\
                    "Dynamic Filter value {} must be either in keys or values".format(option))
            }
        ]

    def get_sort(self):
        sortby = self.options.get("sortby", None)
        if sortby == 'Keys ASC':
            return "sort({})".format(",".join([key+":ascending" for key in self.getKeyFields()]))
        elif sortby == 'Keys DESC':
            return "sort({})".format(",".join([key+":descending" for key in self.getKeyFields()]))
        elif sortby == 'Values ASC':
            return "sort({})".format(",".join([key+":ascending" for key in self.getValueFields()]))
        elif sortby == 'Values DESC':
            return "sort({})".format(",".join([key+":descending" for key in self.getValueFields()]))
        return ""

    def doRenderChart(self):
        payload = self.compute_brunel_magic()
        if not isinstance(payload, tuple):
            payload = (self.getWorkingPandasDataFrame(), payload)
        pandas_df, magic_parts = payload
        ShellAccess['brunel_temp_df'] = pandas_df
        try:
            with capture_output() as buf:
                if self.use_online_js() and not self.is_gateway:
                    ipythonDisplay(HTML("""
                        <link rel="stylesheet" href="https://code.jquery.com/ui/1.12.0/themes/smoothness/jquery-ui.min.css" type="text/css" />
                        <script src="https://code.jquery.com/jquery-3.2.1.js" integrity="sha256-DZAnKJ/6XZ9si04Hgrsxu/8s717jcIzLy3oi35EouyE=" crossorigin="anonymous"></script>
                        <script src="https://code.jquery.com/ui/1.12.1/jquery-ui.js" integrity="sha256-T0Vest3yCU7pafRw9r+settMBX6JkKN06dqBnpQ8d30=" crossorigin="anonymous"></script>
                        <script src="http://requirejs.org/docs/release/2.2.0/minified/require.js" charset="utf-8"></script>
                    """))
                magic = "data('brunel_temp_df') {}".format(self.complete_magic(magic_parts))
                self.debug("Running brunel with magic {}".format(magic))
                data = get_ipython().run_line_magic('brunel', magic)
                if data is not None:
                    ipythonDisplay(data)
            brunel_html = "\n".join([self.convert_html(output) for output in buf.outputs])
            return brunel_html
        finally:
            del ShellAccess['brunel_temp_df']




# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.display.chart.renderers.baseChartDisplay import commonChartOptions
from pixiedust.utils import Logger
from .bokehBaseDisplay import BokehBaseDisplay
from bokeh.plotting import figure
from bokeh.models import HoverTool
import numpy as np
import math
import sys

@PixiedustRenderer(id="histogram")
@Logger()
class BKHistogramRenderer(BokehBaseDisplay):
    def supportsAggregation(self, handlerId):
        return False

    def supportsKeyFields(self, handlerId):
        return False
    
    def getPreferredDefaultValueFieldCount(self, handlerId):
        return 1

    def getDefaultKeyFields(self, handlerId, aggregation):
        return []

    def acceptOption(self, optionName):
        if optionName == 'histoChartType':
            return False
        return True

    @commonChartOptions
    def getChartOptions(self):
        return [
            { 'name': 'color',
              'metadata': {
                    'type': "dropdown",
                    'values': ["None"] + self.getFieldNames(),
                    'default': ""
                }
            }
        ]

    def canRenderChart(self):
        return (True, None)

    def getExtraFields(self):
        color = self.options.get("color")
        return [color] if color is not None else []

    def createBokehChart(self):
        defaultbin = math.sqrt(len(self.getWorkingPandasDataFrame().index))
        binsize = int(self.options.get('binsize', defaultbin))
        valueFields = self.getValueFields()
        colour = self.options.get("color")

        def histogram(df, vField, color=None, clustered=None):
            colors = self.colorPalette(len(df.index)) if color is None else color

            p = figure(y_axis_label='Frequency', x_axis_label=vField)

            for j,c in enumerate(list(df[clustered].unique())) if clustered else enumerate([None]):
                df2 = df[df[clustered] == c] if c else df
                hist, edges = np.histogram(list(df2[vField]), density=True, bins=binsize)
                p.quad(top=hist, bottom=0, left=edges[:-1], right=edges[1:], fill_color=colors[j], line_color="#cccccc", fill_alpha=0.8 if clustered else 1, legend=str(c) if clustered else None)

            p.y_range.start=0

            p.legend.location = "top_left"

            hover = HoverTool()
            hover.tooltips = [('Frequency', '@top{0.00}'), ('Interval', '@left - @right')]
            p.add_tools(hover)

            return p

        charts = []
        wpdf = self.getWorkingPandasDataFrame().copy()
        colours = self.colorPalette(None if colour is None else len(wpdf[colour].unique()))
        
        for i, v in enumerate(valueFields):
            charts.append(histogram(wpdf, v, color=colours, clustered=colour))


        return charts




# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.utils import Logger
from .brunelBaseDisplay import BrunelBaseDisplay

@PixiedustRenderer(id="lineChart")
@Logger()
class LineChartRenderer(BrunelBaseDisplay):

    def isSubplot(self):
        return self.options.get("lineChartType", None) == "subplots"

    def getExtraFields(self):
        if not self.isSubplot() and len(self.getValueFields())>1:
            #no clusterby if we are grouped and multiValueFields
            return []
    
        clusterby = self.options.get("clusterby")
        return [clusterby] if clusterby is not None else []

    def compute_brunel_magic(self):
        parts = ["line"]
        subplots = self.isSubplot()
        valueFields = self.getValueFields()
        keyFields = self.getKeyFields()
        clusterby = self.options.get("clusterby")
        
        if subplots is False:
            if clusterby is None:
                parts.append("x({})".format(keyFields[0]))
                if len(valueFields) == 1:
                    parts.append("y({})".format(valueFields[0]))
                elif len(valueFields) > 1:
                    parts.append("y({})".format(",".join(valueFields)))
                    parts.append("color(#series)")
            elif clusterby is not None:
                if len(valueFields) > 1:
                    self.addMessage("Warning: 'Cluster By' ignored when grouped option with multiple Value Fields is selected") 
        elif subplots is True:
            if clusterby is None:
                parts.append("x({})".format(keyFields[0]))
                parts.append("y({})".format(valueFields[0]))   
                if len(valueFields) > 1:
                    for panel in range(1,len(valueFields)):
                        parts.append("| line x({})".format(keyFields[0]))
                        parts.append("y({})".format(valueFields[panel]))
            elif clusterby is not None:
                self.addMessage("Warning: 'Cluster By' not implemented for Brunel yet")

        parts.append(self.get_sort())

        return parts



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.app import *
from pixiedust.utils import Logger
from .baseOptions import BaseOptions
import re

@Logger()
class ChartOptionTitle(object):
    @route(widget="pdChartOptionTitle")
    def chart_option_title_widget(self, optid, title):
        return """
<label for="chartoption{{optid}}{{prefix}}">Chart Title:</label>
<input type="text" class="form-control" id="chartoption{{optid}}{{prefix}}" name="{{optid}}" value="{{title or ""}}"
    pd_script="self.options_callback('title', '$val(chartoption{{optid}}{{prefix}})')" onkeyup="$(this).trigger('click');">
"""

@PixieApp
@Logger()
class OptionsShell(BaseOptions, ChartOptionTitle):
    def camel_case_to_title(self, name):
        return re.compile('([a-z0-9])([A-Z])').sub(r'\1 \2', name).title()

    def get_custom_options(self):
        "Options for this dialog"
        t = self.camel_case_to_title(self.parsed_command['kwargs']['handlerId']) if 'kwargs' in self.parsed_command and 'handlerId' in self.parsed_command['kwargs'] else 'Chart'
        return {
            "runInDialog":"true",
            "title": t + " Options",
            "showFooter":"true",
            "customClass":"pixiedust-default-dialog"
        }

    def setup(self):
        BaseOptions.setup(self)
        self.chart_options = [{
            "optid": "title",
            "title": lambda: self.run_options.get("title") or "",
            "widget": "pdChartOptionTitle"
        }]
        self.new_options = {}

    def options_callback(self, option, value):
        # self.options[option] = value
        self.new_options[option] = value

    def get_option(self, optid=None, widget=None):
        """
        returns the first chart_option that matches the given arguments
        """
        index = self.get_option_index(optid=optid, widget=widget)
        return None if index == -1 else self.chart_options[index]

    def get_option_index(self, optid=None, widget=None):
        """
        returns the index of the first chart_option that matches the given arguments
        """
        if self.chart_options is None or len(self.chart_options) == 0:
            return -1
        elif optid is None and widget is None:
            return -1
        else:
            option = -1
            for index, chartopt in enumerate(self.chart_options):
                if (optid is None or chartopt['optid'] == optid) and (widget is None or chartopt['widget'] == widget):
                    option = index
                if option != -1:
                    break
            return option

    @route()
    def main_screen(self):
        self.fieldNamesAndTypes = self.get_field_names_and_types(True, True)
        self.fieldNames = self.get_field_names(True)
        self._addHTMLTemplate("optionsshell.html")

    def get_new_options(self):
        return self.new_options



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from abc import abstractmethod, ABCMeta
from six import iteritems
from pixiedust.display.app import *
from pixiedust.utils.astParse import parse_function_call
from pixiedust.utils import cache, Logger
from pixiedust.display.datahandler import getDataHandler
from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.utils.shellAccess import ShellAccess
from IPython.core.getipython import get_ipython

@PixieApp
@Logger()
class BaseOptions(with_metaclass(ABCMeta)):
    def setup(self):
        self.parsed_command = parse_function_call(self.parent_command) #pylint: disable=E1101,W0201
        self.parent_entity = self.parsed_command['args'][0] if len(self.parsed_command['args'])>0 else None
        if self.parent_entity is not None:
            try:
                self.parent_entity = eval(self.parent_entity, get_ipython().user_ns) #pylint: disable=W0123
            except Exception as exc:
                self.exception(exc)
                self.entity = None

        #apply the cell_metadata
        if getattr(self, "cell_metadata"):
            self.parsed_command['kwargs'].update( 
                self.cell_metadata.get("pixiedust",{}).get("displayParams",{})
            )

    @cache(fieldName="fieldNames")
    def get_field_names(self, expandNested=True):
        return self.data_handler.getFieldNames(expandNested)

    @cache(fieldName="fieldNamesAndTypes")
    def get_field_names_and_types(self, expandNested=True, sorted=False):
        return self.data_handler.getFieldNamesAndTypes(expandNested, sorted)

    def get_custom_options(self):
        "Options for this base dialog"
        return {
            "runInDialog":"true",
            "title":"Test",
            "showFooter":"true"
        }

    def on_ok(self, **kwargs):
        run_options = self.run_options
        #update with the new options
        run_options.update(self.get_new_options())
        command = "{}({},{})".format(
            self.parsed_command['func'],
            self.parsed_command['args'][0],
            ",".join(
                ["{}='{}'".format(k, v) for k, v in iteritems(run_options)]
            )
        )
        js = self.env.from_string("""
            pixiedust.executeDisplay(
                {{this.get_pd_controls(
                    command=command, 
                    avoidMetadata=avoid_metadata, 
                    override_keys=override_keys,
                    options=run_options, 
                    black_list=['nostore_figureOnly'],
                    prefix=this.run_options['prefix'],
                ) | tojson}}
            );
        """).render(
            this=self, 
            run_options=run_options, 
            command=command, 
            avoid_metadata=kwargs.get("avoid_metadata", True),
            override_keys=kwargs.get("override_keys", [])
        )
        return self._addJavascript(js)

    @property
    def run_options(self):
        black_list = ['runInDialog','nostore_figureOnly','targetDivId','nostore_cw']
        return {key:value for key,value in iteritems(self.parsed_command['kwargs']) if key not in black_list}

    @property
    def data_handler(self):
        return getDataHandler(
            self.parsed_command['kwargs'], self.parent_entity
        ) if self.parent_entity is not None else None

    @property
    def get_renderer(self):
        return PixiedustRenderer.getRenderer(
            self.parsed_command['kwargs'],
            self.parent_entity, False
        ) if self.parent_entity is not None else None

    @abstractmethod
    def get_new_options(self):
        "return a dictionary of the new options selected by the user"
        pass



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

import sqlite3
import os

import hashlib
import json
import getpass
import sys
import time
from pixiedust.utils.constants import PIXIEDUST_REPO_URL
from pkg_resources import get_distribution
from re import search
from requests import post
from os import environ as env
from pixiedust.utils.printEx import *

from . import pdLogging
logger = pdLogging.getPixiedustLogger()
getLogger = pdLogging.getLogger

myLogger = getLogger(__name__)

__all__ = ['Storage']

SQLITE_DB_NAME = os.environ.get('PIXIEDUST_DB_NAME', 'pixiedust.db')
SQLITE_DB_NAME_PATH = os.environ.get("PIXIEDUST_HOME", os.path.expanduser('~')) + "/" + SQLITE_DB_NAME

if not os.path.exists(os.path.dirname(SQLITE_DB_NAME_PATH)):
    os.makedirs(os.path.dirname(SQLITE_DB_NAME_PATH))

#global connection
_conn = None

def _initStorage():
    def copyRename(oldName, newName):
        if os.path.isfile(oldName) and not os.path.isfile(newName):
            from shutil import copyfile
            copyfile(oldName, newName)
            os.rename(oldName, oldName + ".migrated")
            print("INFO: migrated pixiedust database")

    #if db already exist with old name, rename it now
    copyRename('spark.db', SQLITE_DB_NAME)
    copyRename(SQLITE_DB_NAME, SQLITE_DB_NAME_PATH)

    global _conn
    if not _conn:
        def _row_dict_factory(cursor,row):
            res={}
            for i,col in enumerate(cursor.description):
                res[col[0]]=row[i]
            return res
        _conn = sqlite3.connect(SQLITE_DB_NAME_PATH)
        _conn.row_factory=_row_dict_factory 
        print("Pixiedust database opened successfully")

    _trackDeployment()

"""
Encapsule access to data from the pixiedust database
including storage lifecycle e.g. schema definition and creation, cleanup, etc...
"""
class Storage(object):
    def __init__(self):
        pass

    def _tableExists(self, tableName):
        cursor=_conn.execute("""
            SELECT * FROM sqlite_master WHERE name ='{0}' and type='table';
        """.format(tableName))
        exists = cursor.fetchone() is not None
        cursor.close()
        return exists

    def _initTable(self, tableName, schemaDef):
        cursor=_conn.execute("""
            SELECT * FROM sqlite_master WHERE name ='{0}' and type='table';
        """.format(tableName))
        if cursor.fetchone() is None:
            _conn.execute('''CREATE TABLE {0} ({1});'''.format(tableName, schemaDef))
            print("Table {0} created successfully".format(tableName))
        '''
        else:
            print("Deleting table")
            _conn.execute("""DROP TABLE {0}""".format(tableName))
        '''
        cursor.close()

    def fetchOne(self, sqlQuery, mapper=None):
        cursor=None
        try:
            cursor=_conn.execute(sqlQuery)
            row = cursor.fetchone()
            return row if row is None or mapper is None else mapper(row)
        finally:
            if cursor is not None:
                cursor.close()

    def fetchMany(self, sqlQuery, mapper=None):
        cursor=None
        try:
            cursor=_conn.execute(sqlQuery)
            results = cursor.fetchmany()
            ret = []
            while results:
                for row in results:
                    ret.append( row if mapper is None else mapper(row) )
                results = cursor.fetchmany()
            return ret
        finally:
            if cursor is not None:
                cursor.close()

    def execute(self, sqlQuery, handler):
        cursor=None
        try:
            cursor=_conn.execute(sqlQuery)
            results = cursor.fetchmany()
            while results:
                for row in results:
                    handler(row)
                results = cursor.fetchmany()
        finally:
            if cursor is not None:
                cursor.close()

    """Return the number of rows deleted"""
    def delete(self, sqlQuery):
        cursor = None
        try:
            cursor=_conn.execute(sqlQuery)
            _conn.commit()
            return cursor.rowcount
        finally:
            if cursor is not None:
                cursor.close()

    def insert(self, sqlQuery, arguments=None):
        if arguments is not None:
            _conn.execute(sqlQuery, arguments)
        else:
            _conn.execute(sqlQuery)
        _conn.commit()

    def update(self, sqlQuery, arguments=None):
        if arguments is not None:
            _conn.execute(sqlQuery, arguments)
        else:
            _conn.execute(sqlQuery)
        _conn.commit()

PIXIEDUST_VERSION_TBL_NAME = "VERSION_TRACKER"
METRICS_TRACKER_TBL_NAME = "METRICS_TRACKER"

class __DeploymentTrackerStorage(Storage):
    def __init__(self):
        self._initTable(PIXIEDUST_VERSION_TBL_NAME,"VERSION TEXT NOT NULL")

def _trackDeployment():
    deploymenTrackerStorage = __DeploymentTrackerStorage()
    doNotTrack = None
    lastVersionTracked = None
    # if the metrics tracker table does not exist then this represents a new install
    # we do not want to track the user until after they have a chance to opt out
    if not deploymenTrackerStorage._tableExists(METRICS_TRACKER_TBL_NAME):
        doNotTrack = True
        deploymenTrackerStorage._initTable(METRICS_TRACKER_TBL_NAME,"LAST_VERSION_TRACKED TEXT NULL, OPT_OUT BOOLEAN NOT NULL")
        deploymenTrackerStorage.insert("INSERT INTO {0} (OPT_OUT) VALUES ({1})".format(METRICS_TRACKER_TBL_NAME,0))
        print("""
Share anonymous install statistics? (opt-out instructions)

PixieDust will record metadata on its environment the next time the package is installed or updated. The data is anonymized and aggregated to help plan for future releases, and records only the following values:

{
   "data_sent": currentDate,
   "runtime": "python",
   "application_version": currentPixiedustVersion,
   "space_id": nonIdentifyingUniqueId,
   "config": {
       "repository_id": "https://github.com/ibm-watson-data-lab/pixiedust",
       "target_runtimes": ["Data Science Experience"],
       "event_id": "web",
       "event_organizer": "dev-journeys"
   }
}
You can opt out by calling pixiedust.optOut() in a new cell.""")
    else:
        row = deploymenTrackerStorage.fetchOne("SELECT * FROM {0}".format(METRICS_TRACKER_TBL_NAME));
        if row is None:
            deploymenTrackerStorage.insert("INSERT INTO {0} (OPT_OUT) VALUES ({2})".format(METRICS_TRACKER_TBL_NAME,0))
            doNotTrack = False
            lastVersionTracked = None
        else:
            doNotTrack = row["OPT_OUT"] == 1
            lastVersionTracked = row["LAST_VERSION_TRACKED"]

    row = deploymenTrackerStorage.fetchOne("SELECT * FROM {0}".format(PIXIEDUST_VERSION_TBL_NAME));
    if row is None:
        _updateVersionAndTrackDeployment(deploymenTrackerStorage, None, lastVersionTracked, doNotTrack)
    else:
        _updateVersionAndTrackDeployment(deploymenTrackerStorage, row["VERSION"], lastVersionTracked, doNotTrack)

def _updateVersionAndTrackDeployment(deploymenTrackerStorage, lastPixiedustVersion, lastVersionTracked, doNotTrack):
    # Get version and repository URL from 'setup.py'
    version = None
    try:
        app = get_distribution("pixiedust")
        version = app.version
        # save last tracked version in the db
        if lastPixiedustVersion is None:
            deploymenTrackerStorage.insert("INSERT INTO {0} (VERSION) VALUES ('{1}')".format(PIXIEDUST_VERSION_TBL_NAME,version))
        else:
            deploymenTrackerStorage.update("UPDATE {0} SET VERSION='{1}'".format(PIXIEDUST_VERSION_TBL_NAME,version))
        # if version has changed then track with deployment tracker
        if lastPixiedustVersion is None or lastPixiedustVersion != version:
            myLogger.info("Change in version detected: {0} -> {1}.".format(lastPixiedustVersion,version))
            if lastPixiedustVersion is None:
                printWithLogo("Pixiedust version {0}".format(version))
            else:
                printWithLogo("Pixiedust version upgraded from {0} to {1}".format(lastPixiedustVersion,version))
            # register
            if not doNotTrack:
                deploymenTrackerStorage.update("UPDATE {0} SET LAST_VERSION_TRACKED='{1}'".format(METRICS_TRACKER_TBL_NAME,version))
                track(version)
        else:
            myLogger.info("No change in version: {0} -> {1}.".format(lastPixiedustVersion,version))
            printWithLogo("Pixiedust version {0}".format(version))
            # if never tracked then track for the first time
            if lastVersionTracked is None and not doNotTrack:
                deploymenTrackerStorage.update("UPDATE {0} SET LAST_VERSION_TRACKED='{1}'".format(METRICS_TRACKER_TBL_NAME,version))
                track(version)
    except:
        myLogger.error("Error registering with deployment tracker:\n" + str(sys.exc_info()[0]) + "\n" + str(sys.exc_info()[1]))

def track(version):
    event = dict()
    event['date_sent'] = time.strftime("%Y-%m-%dT%H:%M:%S.000Z", time.gmtime())
    event['runtime'] = 'python'
    if version is not None:
        event['application_version'] = version
    try:
        # a hashed value that uniquely identifies this user
        # no password or identifying information are tracked
        event['space_id'] = hashlib.md5(getpass.getuser()).hexdigest()
    except:
        pass
    event['config'] = {}
    event['config']['repository_id'] = PIXIEDUST_REPO_URL
    event['config']['target_runtimes'] = ['Data Science Experience']
    event['config']['event_id'] = 'web'
    event['config']['event_organizer'] = 'dev-journeys'
    url = 'https://metrics-tracker.mybluemix.net/api/v1/track'
    headers = {'content-type': "application/json"}
    try:
        response = post(url, data=json.dumps(event), headers=headers)
        myLogger.info('Anonymous install statistics collected.')
    except Exception as e:
        myLogger.error('Anonymous install statistics collection error: %s' % str(e))

def optOut():
    deploymenTrackerStorage = __DeploymentTrackerStorage()
    deploymenTrackerStorage.update("UPDATE {0} SET OPT_OUT={1}".format(METRICS_TRACKER_TBL_NAME,1))
    print("Pixiedust will not collect anonymous install statistics.")

def optIn():
    deploymenTrackerStorage = __DeploymentTrackerStorage()
    deploymenTrackerStorage.update("UPDATE {0} SET OPT_OUT={1}".format(METRICS_TRACKER_TBL_NAME,0))
    print("Pixiedust will collect anonymous install statistics. To opt out, call pixiedust.optOut() in a new cell.")


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from pixiedust.utils.environment import Environment
from pixiedust.display.streaming import StreamingDataAdapter
import pixiedust.utils.dataFrameMisc as dataFrameMisc
from .jsonDataHandler import JSONDataHandler
from .pandasDataFrameHandler import PandasDataFrameDataHandler
if Environment.hasSpark:
    from .pysparkDataFrameHandler import PySparkDataFrameDataHandler

def isArrayOfDict(entity):
    if not isinstance(entity, list):
        return False
    return all([isinstance(x, dict) for x in entity])

def getDataHandler(options, entity):
    if dataFrameMisc.isPySparkDataFrame(entity):
        return PySparkDataFrameDataHandler(options, entity)
    elif dataFrameMisc.isPandasDataFrame(entity):
        return PandasDataFrameDataHandler(options, entity)
    elif isinstance(entity, dict) or isArrayOfDict(entity):
        return JSONDataHandler(options, entity)
    elif isinstance(entity, StreamingDataAdapter):
        return entity.getDisplayDataHandler(options, entity)

    return None


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from pixiedust.utils.dataFrameAdapter import PandasDataFrameAdapter
import pixiedust.utils.dataFrameMisc as dataFrameMisc
from datetime import datetime
import pandas as pd
import numpy as np
from pixiedust.utils import Logger
from six import iteritems
from .baseDataHandler import BaseDataHandler
import re

@Logger()
class PandasDataFrameDataHandler(BaseDataHandler):

    def getFieldNames(self, expandNested=False):
        return dataFrameMisc.getFieldNames(PandasDataFrameAdapter(self.entity), expandNested)

    def count(self):
        return len(self.entity.index)

    def isNumericField(self, fieldName):
        for y in self.entity.columns:
            if y == fieldName:
                return self.entity[y].dtype == np.float64 or self.entity[y].dtype == np.int64
        raise ValueError("Column {} does not existing in the dataframe".format(fieldName))

    def isNumericType(self, field):
        return dataFrameMisc.isNumericType(field.dataType)

    def isStringField(self, fieldName):
        return dataFrameMisc.isStringField(PandasDataFrameAdapter(self.entity), fieldName)

    def isStringType(self, field):
        return dataFrameMisc.isStringType(field.dataType)

    def isDateField(self, fieldName):
        return dataFrameMisc.isDateField(PandasDataFrameAdapter(self.entity), fieldName)

    def isDateType(self, field):
        return dataFrameMisc.isDateType(field.dataType)

    @property
    def schema(self):
        return PandasDataFrameAdapter(self.entity).schema

    def add_numerical_column(self):
        """
        Add a dummy numerical column to the underlying dataframe
        """
        self.entity = self.entity.copy()
        self.entity["pd_count"] = 1
        return "pd_count"

    def get_filtered_dataframe(self, filter_options):
        df = self.entity.copy(deep=True)

        if filter_options is not None:
            field = filter_options['field'] if 'field' in filter_options else ''
            constraint = filter_options['constraint'] if 'constraint' in filter_options else ''
            val = filter_options['value'] if 'value' in filter_options else ''
            regex = filter_options['regex'].lower() == "true" if 'regex' in filter_options else False
            casematters = filter_options['case_matter'].lower() == "true" if 'case_matter' in filter_options else False

            if field and val and field in self.getFieldNames():
                if val == 'None':
                    df = df.loc[df[field].isna() if hasattr(df[field], 'isna') else df[field].isnull()]
                elif not self.isNumericField(field):
                    val = val if regex else ".*" + val + ".*"
                    flags = 0 if casematters else re.IGNORECASE
                    df = df[df[field].str.contains(val, flags=flags)]
                elif constraint == "less_than":
                    df = df.loc[df[field] < float(val)]
                elif constraint == "greater_than":
                    df = df.loc[df[field] > float(val)]
                else: # constraint == "equal_to":
                    df = df.loc[df[field] == float(val)]
        return df

    """
        Return a cleaned up Pandas Dataframe that will be used as working input to the chart
    """
    def getWorkingPandasDataFrame(self, xFields, yFields, extraFields=[], aggregation=None, maxRows = 100, filterOptions={}, isTableRenderer=False):
        filteredDF = self.get_filtered_dataframe(filterOptions)

        if xFields is None or len(xFields)==0:
            #swap the yFields with xFields
            xFields = yFields
            yFields = []
            aggregation = None

        allFields = self.getFieldNames()
        myFieldsOrdered = []
        if isTableRenderer:
            if len(extraFields) < 1:
                workingDF = filteredDF
            else:
                for f in allFields:
                    if f in extraFields:
                        myFieldsOrdered.append(f)
                workingDF = filteredDF[myFieldsOrdered]
        else:
            extraFields = [a for a in extraFields if a not in xFields and a not in yFields]
            workingDF = filteredDF[xFields + extraFields + yFields]

        if aggregation and len(yFields)>0:
            aggMapper = {"SUM":"sum", "AVG": "mean", "MIN": "min", "MAX": "max"}
            aggFn = aggMapper.get(aggregation, "count")
            workingDF = workingDF.groupby(extraFields + xFields).agg(aggFn).reset_index()

        if not isTableRenderer:
            workingDF = workingDF.dropna()
        count = len(workingDF.index)
        if count > maxRows:
            workingDF = workingDF.sample(n=int(maxRows),replace=False)

        #check if the caller want to preserve some columns
        preserveCols = self.options.get("preserveCols", None)
        if preserveCols is not None:
            preserveCols = [a for a in preserveCols.split(",") if a not in xFields and a not in yFields]
            cols = { key:[] for key in preserveCols}
            for i in workingDF.index:
                cond = None
                for j, key in enumerate(extraFields + xFields):
                    thisKey = self.entity.loc[i][key]
                    cond = (self.entity[key] == thisKey) if j is 0 else (cond & (self.entity[key] == thisKey))
                p = self.entity[ cond ]
                for key in preserveCols:
                    cols[key].append( p[key].values.tolist()[0] )
            for key, value in iteritems(cols):
                workingDF.insert(len(workingDF.columns), key, value)

        #check if the user wants timeseries
        if len(xFields) == 1 and self.options.get("timeseries", 'false') == 'true':
            field = xFields[0]
            try:
                inputDateFormat = self.options.get("inputDateFormat", None)
                if inputDateFormat is not None:
                    workingDF[field] = workingDF[field].apply(lambda x: datetime.strptime(str(x), inputDateFormat))
                else:
                    workingDF[field] = pd.to_datetime(workingDF[field])
            except:
                self.exception("Unable to convert field {} to datetime".format(field))
        
        if not isTableRenderer:
            #sort by xFields
            workingDF.sort_values(extraFields + xFields, inplace=True)
        
        return workingDF


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

import math
from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.utils import Logger
from .brunelBaseDisplay import BrunelBaseDisplay

@PixiedustRenderer(id="histogram")
@Logger()
class HistogramRenderer(BrunelBaseDisplay):
    def supportsAggregation(self, handlerId):
        return False

    def supportsKeyFields(self, handlerId):
        return False
    
    def getPreferredDefaultValueFieldCount(self, handlerId):
        return 1

    def getDefaultKeyFields(self, handlerId, aggregation):
        return []

    def acceptOption(self, optionName):
        if optionName == 'histoChartType':
            return False
        return True

    def compute_brunel_magic(self):
        parts = ["bar"]
        defaultbin = math.sqrt(len(self.getWorkingPandasDataFrame().index))
        binsize = int(self.options.get('binsize', defaultbin))

        for index, key in enumerate(self.getValueFields()):
            if index > 0:
                parts.append("+ bar")
            parts.append("x({0}) bin({0}:{1})".format(key, binsize))
            parts.append("y(#count)")
            parts.append("""style("size:100%")""")
        return parts



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from .bokehBaseDisplay import BokehBaseDisplay
from pixiedust.utils import Logger
import numpy as np
from bokeh.plotting import figure
from bokeh.models import ColumnDataSource, FactorRange, HoverTool
from bokeh.core.properties import value as bk_value
from bokeh.transform import factor_cmap
import sys


@PixiedustRenderer(id="barChart")
@Logger()
class BKBarChartRenderer(BokehBaseDisplay):
    def getSubplotHSpace(self):
        return 0.5

    def isSubplot(self):
        return self.options.get("charttype", "grouped") == "subplots"

    def getExtraFields(self):
        if not self.isSubplot() and len(self.getValueFields())>1:
            #no clusterby if we are grouped and multiValueFields
            return []

        clusterby = self.options.get("clusterby")
        return [clusterby] if clusterby is not None else []

    def createBokehChart(self):
        keyFields = self.getKeyFields()
        valueFields = self.getValueFields()
        clusterby = self.options.get("clusterby")
        stacked = self.options.get("charttype", "grouped") == "stacked"
        multivalue =  len(valueFields) > 1

        def convertPDFDate(df, col):
            #Bokeh doesn't support datetime as index in Bar chart. Convert to String
            if len(keyFields) == 1:
                dtype = df[col].dtype.type if col in df else None
                if np.issubdtype(dtype, np.datetime64):
                    dateFormat = self.options.get("dateFormat", None)
                    try:
                        df[col] = df[col].apply(lambda x: str(x).replace(':','-') if dateFormat is None else x.strftime(dateFormat))
                    except:
                        self.exception("Error converting dateFormat {}".format(dateFormat))
                        df[col] = df[col].apply(lambda x: str(x).replace(':','-'))

        def stackedBar(df, xlabel, vFields, color=None, clustered=False, title=None):
            ylabel = ','.join(v for v in vFields)
            ystart = 0

            if clustered:
                factors=list(df.index)
                l = [ bk_value(a) for a in list(df.index) ]
                data = {'pd_stacked_col': list(df.columns.values)}
                for x in list(df.index):
                    data[x] = list(df[x:x].values[0])
                    ystart = min(ystart, min(data[x]))
            else:
                factors=vFields
                l = [ bk_value(a) for a in vFields ]
                data = {'pd_stacked_col': list(df[xlabel].values)}
                for v in vFields:
                    data[v] = list(df[v].values)
                    ystart = min(ystart, min(data[v]))
            
            src = ColumnDataSource(data)
            colors = self.colorPalette(len(factors)) if color is None else color

            p = figure(x_range=data['pd_stacked_col'], y_axis_label=ylabel, x_axis_label=xlabel, title=title)
            p.vbar_stack(factors, x='pd_stacked_col', width=0.9, source=src, legend=l if self.showLegend() else None, color=colors)

            p.y_range.start = ystart
            p.axis.minor_tick_line_color = None
            p.outline_line_color = None
            p.x_range.range_padding = 0.1
            p.xaxis.major_label_orientation = 1
            p.xgrid.grid_line_color = None
            p.legend.location = "top_left"

            hover = HoverTool()
            hover.tooltips = [(d if d is not 'pd_stacked_col' else xlabel, '@' + d + '{0.00}') for d in data]
            p.add_tools(hover)

            return p

        def groupedBar(df, xlabel, vFields, color=None, clustered=False, title=None):
            ylabel = ','.join(v for v in vFields)

            if clustered:
                factors=list(df.index)
                x = [ (b, a) for b in list(df.columns.values) for a in list(df.index) ]
                l = [ (a) for b in list(df.columns.values) for a in list(df.index) ]
                counts = sum(zip(df.at[a,b] for b in list(df.columns.values) for a in list(df.index)), ())
            else:
                factors=vFields
                x = [ (b,a) for b in list(df[xlabel].values) for a in vFields ]
                l = [ (a) for b in list(df[xlabel].values) for a in vFields ]
                counts = [ df[df[xlabel] == b][a].values[0] for b in list(df[xlabel].values) for a in vFields ]

            src = ColumnDataSource(data=dict(x=x, counts=counts, l=l))
            colors = self.colorPalette(len(factors)) if color is None else color

            p = figure(x_range=FactorRange(*x), y_axis_label=ylabel, x_axis_label=xlabel, title=title)
            p.vbar(x='x', top='counts', width=0.925, source=src, legend='l' if self.showLegend() else None, color=factor_cmap('x', palette=colors, factors=factors, start=1, end=2))

            p.y_range.start = 0 if not counts else min(0, min(counts))
            p.axis.minor_tick_line_color = None
            p.outline_line_color = None
            p.x_range.range_padding = 0.1
            p.xaxis.major_label_orientation = 1
            p.xaxis.major_label_text_font_size = "0px"
            p.xaxis.major_label_text_color = None
            p.xaxis.major_tick_line_color = None
            p.xgrid.grid_line_color = None
            p.legend.location = "top_left"

            hover = HoverTool()
            hover.tooltips = [(xlabel, '@x'), (ylabel, '@counts{0.00}')]
            p.add_tools(hover)

            return p

        charts = []
        wpdf = self.getWorkingPandasDataFrame().copy()
        xlabel = ','.join(k for k in keyFields)

        for index, row in wpdf.iterrows():
            for k in keyFields:
                if isinstance(row[k], str if sys.version >= '3' else basestring):
                    row[k] = row[k].replace(':', '.')
            wpdf.loc[index] = row

        if len(keyFields) > 1:
            wpdf[xlabel] = wpdf[keyFields].apply(lambda x: ','.join(str(y) for y in x), axis=1)
        else:
            wpdf[xlabel] = wpdf[xlabel].astype(str)

        if self.isSubplot():
            colors = self.colorPalette(len(valueFields) * (len(wpdf[clusterby].unique())) if clusterby else len(valueFields))
            k = 0
            for i, valueField in enumerate(valueFields):
                for j,c in enumerate(list(wpdf[clusterby].unique())) if clusterby else enumerate([None]):
                    wpdf2 = wpdf[wpdf[clusterby] == c] if c else wpdf
                    wpdf2 = wpdf2.drop(clusterby, axis=1) if c else wpdf2
                    charts.append(stackedBar(wpdf2, xlabel, [valueFields[i]], clustered=False, color=colors[k], title="{} = {}".format(clusterby, c) if c else None))
                    k += 1

        elif clusterby and not multivalue:
            ylabel = ','.join(v for v in valueFields)
            wpdf[clusterby] = wpdf[clusterby].astype(str)
            wpdf = wpdf.pivot(clusterby, xlabel, ylabel)
            wpdf = wpdf.fillna(0)

            if stacked:
                charts.append(stackedBar(wpdf, xlabel, valueFields, clustered=True))
            else:
                charts.append(groupedBar(wpdf, xlabel, valueFields, clustered=True))

        else:
            if stacked:
                charts.append(stackedBar(wpdf, xlabel, valueFields, clustered=False))
            else:
                charts.append(groupedBar(wpdf, xlabel, valueFields, clustered=False))

            if clusterby:
                self.addMessage("Warning: 'Cluster By' ignored when grouped option with multiple Value Fields is selected")

        return charts



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

class BaseDataHandler(object):
    def __init__(self, options, entity):
        self.options = options
        self.entity = entity
        self.isStreaming = False

    def add_numerical_column(self):
        raise NotImplementedError()

    def count(self):
        raise NotImplementedError()

    def getFieldNamesAndTypes(self, expandNested=True, sorted=False):
        fieldNames = self.getFieldNames(expandNested)
        fieldNamesAndTypes = []
        for fieldName in fieldNames:
            fieldType = "unknown/unsupported"
            if self.isNumericField(fieldName):
                fieldType = "numeric"
            elif self.isDateField(fieldName):
                fieldType = "date/time"
            elif self.isStringField(fieldName):
                fieldType = "string"
            fieldNamesAndTypes.append((fieldName, fieldType))
        if sorted:
            fieldNamesAndTypes.sort(key=lambda x: str(x[0]))
        return fieldNamesAndTypes

    def get_filtered_dataframe(self, filter_options):
        return self.entity



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2018
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.app import route
from pixiedust.utils import Logger
import six

@Logger()
class KeyValueSelector(object):
    @route(widget="pdKeyValueSelector")
    def chart_option_key_value_widget(self, optid, keyFields, valueFields):
        self.keyFields = keyFields.split(",") if isinstance(keyFields, six.string_types) else keyFields or []
        self.valueFields = valueFields.split(",") if isinstance(valueFields, six.string_types) and valueFields else valueFields or []
        self.keyFieldsSupported = self.key_fields_supported()
        self.keyFieldsType = self.key_fields_type()
        self.valueFieldsType = self.value_fields_type()
        self._addHTMLTemplate("keyvalueselector.html")

    def key_fields_supported(self):
        return self.get_renderer.supportsKeyFields(self.parsed_command['kwargs']['handlerId'])

    def key_fields_type(self):
        return ['string', 'numeric', 'date/time']

    def value_fields_type(self):
        return ['numeric']



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.utils import Logger
from .brunelBaseDisplay import BrunelBaseDisplay

@PixiedustRenderer(id="pieChart")
@Logger()
class PieChartRenderer(BrunelBaseDisplay):
    def compute_brunel_magic(self):
        parts = ["stack polar bar"]

        for index, key in enumerate(self.getKeyFields()):
            if index > 0:
                parts.append("+ line")
            parts.append("""x("const")""")
            parts.append("y(#count)")
            parts.append("color({0})".format(key))
            parts.append("legends(none)")
            parts.append("label({})".format(key))

        return parts



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from pixiedust.display.display import *
from pixiedust.display import addDisplayRunListener

#add a display Run Listener 
addDisplayRunListener( lambda entity, options: onNewDisplayRun(entity, options) )

activesStreamingEntities = {}
def onNewDisplayRun(entity, options):
    if "cell_id" in options and "showchrome" in options:
        if options["cell_id"] in activesStreamingEntities:
            del activesStreamingEntities[options["cell_id"]]

class StreamingDisplay(Display):
    def __init__(self, options, entity, dataHandler=None):
        super(StreamingDisplay,self).__init__(options,entity,dataHandler)
        self.windowSize = 100